//----------------------------------------------------------------------------
//
// C++ Objects for Allegro's gui
//
// Douglas Eleveld (D.J.Eleveld@anest.azg.nl)
//
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
#include "degui.h"
#include "build_di.h"
#include "build_mu.h"
#include "degui_c.h"
#include <stdio.h>
#include <string.h>

//----------------------------------------------------------------------------
// Graphics settings
int driver = GFX_AUTODETECT;
int xres = 640;
int yres = 480;

int old_driver = GFX_AUTODETECT;
int old_xres = 640;
int old_yres = 480;

//----------------------------------------------------------------------------
// Helper stuff for the builder dialog object managing class
//----------------------------------------------------------------------------
// The header for the dialog files
static char *dialog_file_header = "Allegro gui builder dialog file\n";

// Helper function for object color adjustment
void adjust_object_color (int *color, const char *title);

// Helper function for deleting objects.  This is temporarily set in
// the place of an object to be removed from the dialog.  When the
// dialog is restarted, all DIALOGS pointing to this empty_proc are
// removed from the dialog.  Bit of a hack, but I couldn't figure out
// how to remove an object from a dialog on the fly since it crashed
// when I tried to do it the simple way
int empty_proc (int, DIALOG*, int) { return D_O_K; };

// The main dispatching function i.e. the interface between the C and
// C++ object only for builder objects.  This is special from the
// normal object message dispatching because object in the user's
// dialog are not allowed to actually stop the dialog so we trap the
// close request and give a warning the the dialog would be closed
int builder_object_message_proc (int msg, DIALOG *d, int c)
   {
   // Do the normal message processing
   int ret = dialog_object::object_message_proc(msg,d,c);

   // If the builder object asks for a close just give an alert
   if(ret&D_CLOSE)
      {
      alert(NULL,"The dialog would now be closed",NULL,"&Ok",NULL,'o',0);

      // Reset any closing commands
      ret &= ~D_CLOSE;
      }
   return ret;
   }

//----------------------------------------------------------------------------
// Allegro dialog managing class
//----------------------------------------------------------------------------
// The main dialog
builder_dialog *gui_builder;

//----------------------------------------------------------------------------
builder_dialog::builder_dialog (void)
   :move_key(0),
   resize_key(0),
   pushback_key(0),
   pullfront_key(0),
   delete_key(0),
   fg_color_key(0),
   bg_color_key(0),
   granularity(5),
   res_c(-1),
   res_w(0),
   res_h(0)
   {
   // Setup the menu stuff
   file_menu.add("&Load",load_dialog);
   file_menu.add("&Append",append_dialog);
   file_menu.add("&Save",save_dialog);
   file_menu.add("Ex&port",dialog_export);
   file_menu.add();
   file_menu.add("E&xit",main_dialog_close);

   object_menu.add("&Box",::add_box_object);
   object_menu.add("&Shadow Box",::add_shadow_box_object);
   object_menu.add("B&itmap",::add_bitmap_object);
   object_menu.add("&Text",::add_text_object);
   object_menu.add("&Centerd text",::add_centerd_text_object);
   object_menu.add("&Editable text",::add_edittext_object);
   object_menu.add("B&utton",::add_button_object);
   object_menu.add("Chec&kbox",::add_checkbox_object);
   object_menu.add("&Radio button",::add_radio_button_object);
   object_menu.add("&List Box",::add_list_object);
   object_menu.add("Text B&ox",::add_textbox_object);
   object_menu.add("User",::add_user_object);

   options_menu.add("&Foreground color",::set_fg_color);
   options_menu.add("&Midground color",::set_mg_color);
   options_menu.add("&Background color",::set_bg_color);
   options_menu.add("&Palette",::set_builder_palette);
   options_menu.add("");
   options_menu.add("&Center",::center_dialog);
   options_menu.add("&Granularity",::placement_granularity);
   options_menu.add("&Resolution",::choose_resolution);

   help_menu.add("&Help",::help_procedure);
   help_menu.add("");
   help_menu.add("&About",::about_procedure);

   allegro_menu.add("&File",file_menu);
   allegro_menu.add("&Object",object_menu);
   allegro_menu.add("O&ptions",options_menu);
   allegro_menu.add("H&elp",help_menu);

   // Allocate the necessary control dialog objects
   grid = new grid_object(granularity,granularity);
   if(grid==NULL) degui_no_memory();
   menu_object *main_menu = new menu_object(allegro_menu);
   if(main_menu==NULL) degui_no_memory();

   // Default colours
   RGB fore = {   0/4,  0/4,  0/4 };
   RGB mid  = { 192/4,192/4,192/4 };
   RGB back = { 255/4,255/4,255/4 };

   // Setup the base colours
   ::set_palette(desktop_palette);
   set_color(gui_fg_color,&fore);
   set_color(gui_bg_color,&back);
   set_color(gui_mg_color,&mid);
   use_palette(desktop_palette);

   // Add the necessary control objects
   // Object 0 is added in the dialog constructor
   dialog::add(builder_dialog_proc,0,0,0,0,0,0,0,0,0,0,this); // Object 1
   dialog::add(*grid);                                        // Object 2
   dialog::add(*main_menu);                                   // Object 3
   dialog::add(object_move_proc,0,0,0,0,0,0,move_key,0,granularity,0,NULL);      // Object 4
   dialog::add(object_resize_proc,0,0,0,0,0,0,resize_key,0,granularity,0,NULL);  // Object 5
   }
//----------------------------------------------------------------------------
builder_dialog::~builder_dialog (void)
   {
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   // Delete the all the C++ objects NOTE: even the internal ones
   for(int i=0;i<number;i++)
      {
      // Only delete objects added by us
      if(list[i].proc==builder_object_message_proc)
         {
         // Get the object
         dialog_object *object = (dialog_object*)list[i].dp;
         if(object!=NULL)
            {
            // Get the description
            object_description *descript = (object_description*)object->dataptr;

            // Delete the description if there was one
            if(descript!=NULL) delete descript;

            // Give back the d pointer and any init stuff
            object->disconnect();

            // Delete the object
            delete object;
            }
         }
      }
   if(list!=NULL) delete [] list;
   }
//----------------------------------------------------------------------------
// Add an object to the dialog
void builder_dialog::add (dialog_object *obj)
	{
   // Add the object to the base dialog
   dialog::add(*obj);

	// Make sure the object points to the right message passing procedure
	new_object_dialogs[number-1].proc = builder_object_message_proc;

   // The description has to have been added before we set the index
   if(obj->dataptr!=NULL)
      {
      // Get the description
      object_description *desc = (object_description*)obj->dataptr;

      // Set the index
      desc->set_index(number-5);
      }
	}
//----------------------------------------------------------------------------
// Do the dialog
void builder_dialog::execute (void)
	{
   // Prevent reentrancy
	if(executing==true) return;
   executing = true;

   // Make sure that we have some object_dialogs
   if(number<1) return;

	// Just ask allegro to do the dialog
   do {
      restart = false;

      // Get the newest object list if there is one
      if(new_object_dialogs!=NULL) object_dialogs = new_object_dialogs;
      new_object_dialogs = NULL;

      // Make sure that we our internal objects and procedures are setup
      object_dialogs[1].dp = this;            // General control
      object_dialogs[4].key = move_key;       // For moving
      object_dialogs[4].d1 = granularity;
      object_dialogs[4].dp = object_dialogs;
	   object_dialogs[5].key = resize_key;     // For resizing
      object_dialogs[5].d1 = granularity;
      object_dialogs[5].dp = object_dialogs;

      // Ask Allegro to do the dialog
      int ret = do_dialog(object_dialogs,-1);

      // Don't let an escape exit the gui_builder
      if(ret==-1) reset();

      // If we are restarting for the purpose of deleteing an object
      // then remove all the stuff and put everything back that was
      // not a deleted object
      if(restart==true)
         {
         // Change the resolution here if necessary
         if((res_c>=0)&&(res_w>0)&&(res_h>0))
            {
            // Save the palette for restoration
            PALETTE pal;
            get_palette(pal);

            // Set up the graphics mode
            if(set_gfx_mode(res_c,res_w,res_h,0,0)!=-1)
               {
               old_driver = driver;
               old_xres = xres;
               old_yres = yres;

               driver = res_c;
               xres = res_w;
               yres = res_h;
               }
            // Mode could not be set
            else
               {
               // Try previous mode
               if(set_gfx_mode(old_driver,old_xres,old_yres,0,0)!=-1)
                  {
                  set_palette(pal);
                  alert("Could not set requested video mode",NULL,NULL,"Ok",NULL,13,0);

                  driver = old_driver;
                  xres = old_xres;
                  yres = old_yres;
                  }
               // Previous mode didn't work, I think we have big problems
               else
                  {
                  set_palette(pal);
                  set_gfx_mode(GFX_TEXT,0,0,0,0);
                  fprintf(stderr,"Big problems setting video mode");
                  exit(1);
                  }
               }

            // Restore the palette
            set_palette(pal);

            // Remember that we have changed the video mode
            res_c = -1;
            }
         // Check for added objects
         if(new_object_dialogs!=NULL)
            {
            // Look at all the object_dialogs
            int old_number = number;
            number = 0;

            // Take over the object and dialog pointers
            DIALOG *temp_object_dialogs = new_object_dialogs;
            new_object_dialogs = NULL;

            // Add some of the old object back in
            for(int i=0;i<old_number;i++)
               {
               // Only re add non-empty object_dialogs
               if(temp_object_dialogs[i].proc!=empty_proc)
                  {
                  dialog::add(temp_object_dialogs[i].proc,temp_object_dialogs[i].x,temp_object_dialogs[i].y,temp_object_dialogs[i].w,temp_object_dialogs[i].h,temp_object_dialogs[i].fg,temp_object_dialogs[i].bg,temp_object_dialogs[i].key,temp_object_dialogs[i].flags,temp_object_dialogs[i].d1,temp_object_dialogs[i].d2,temp_object_dialogs[i].dp);
                  }
               // Delete object and description previously added
               else
                  {
                  // Find the object and description behind the DIALOG
                  dialog_object *object = (dialog_object*)temp_object_dialogs[i].dp;
                  if(object!=NULL)
                     {
                     object_description *descript = (object_description*)object->dataptr;
                     if(descript!=NULL) delete descript;
                     delete object;
                     }
                  }
               }
            // All the non-empty object_dialogs have been re added i.e. they are in
            // the new_object_dialogs array so we can delete the object_dialogs and when we
            // go around this while loop the correct object list will be given
            // to the do_dialog function
            delete [] temp_object_dialogs;
            }
         }
      }
      // If an object caused a restart, then do the dialog again
      while(restart==true);

   // Cleanup the stuff
   new_object_dialogs = object_dialogs;
   object_dialogs = NULL;

   // Return the dialog that exited
   executing = false;
	}

//----------------------------------------------------------------------------
// Set some info about the dialog working
void builder_dialog::set_move_key (const int key)
	{
	move_key = key;
   reset();
	}
void builder_dialog::set_resize_key (const int key)
	{
	resize_key = key;
   reset();
	}
void builder_dialog::set_pushback_key (const int key)
	{
	pushback_key = key;
   reset();
	}
void builder_dialog::set_pullfront_key (const int key)
	{
	pullfront_key = key;
   reset();
	}
void builder_dialog::set_delete_key (const int key)
	{
	delete_key = key;
   reset();
	}
void builder_dialog::set_fg_color_key (const int key)
	{
	fg_color_key = key;
   reset();
	}
void builder_dialog::set_bg_color_key (const int key)
	{
   bg_color_key = key;
   reset();
	}
void builder_dialog::set_granularity (const int gran)
	{
	granularity = gran;
   grid->set_granularity(gran,gran);
   reset();
	}
//----------------------------------------------------------------------------
// An Allegro object to get basic messages for the builder
// This is a C++ static function
int builder_dialog::builder_dialog_proc (int msg, DIALOG *d, int c)
    {
    // Call the normal dialog stuff first
    int ret = dialog::dialog_proc(msg,d,c);

    // Get the dialog object
    builder_dialog *dia = (builder_dialog*)d->dp;

    // Check for a restart and exit this dialog so we can start another one
    if(dia->restart==true)
       {
       ret |= D_CLOSE;
       return ret;
       }

    // Look at the message that we were sent
    switch(msg)
       {
       case(MSG_XCHAR):
          {
          // Check for an alt-x for an exit
          if(((c>>8)==KEY_X)&&(key_shifts&=KB_ALT_FLAG))
             {
             // Make sure we won't restart
             dia->restart = false;
             ret |= D_CLOSE;
             }
          // Check for the keycode for a delete
          if((c>>8)==dia->delete_key)
             {
             // Get the Allegro dialogs
             DIALOG *current = dia->object_dialogs;
             DIALOG *pointed = NULL;

             // Find the smallest object being pointed to
             pointed = find_mouse_object(current);

             // Make sure that we pointed to an object
             if(pointed==NULL) return D_O_K;

             // Find the index of the pointed to dialog
             int i;
             int index = -1;
             for(i=0;current[i].proc;i++)
                {
                if(pointed==&(current[i])) index = i;
                }
                // Make sure it was really in there
             if(current[index].proc==NULL) return D_O_K;;

             // Make sure it was in the range that we are allowed to delete
             if((index<6)||(index>=dia->number)) return D_O_K;;

             // Make sure the user wants to delete the object
             if(alert("Delete object?",NULL,NULL,"&Ok","&Cancel",0,0)!=1) return D_O_K;

             // Make the object an empty one
             // When the dialog object is reset the object should be removed
             current[index].proc = empty_proc;

             // Redraw and say that we used the character
             ret = D_REDRAW | D_USED_CHAR;

             // Tell the dialog to restart itself and it should automatically
             // remove any object_dialogs with references to the empty_proc
             dia->reset();
             }
          // Check for an object pushback
          else if((c>>8)==dia->pushback_key)
             {
             // Get the Allegro dialogs
             DIALOG *current = dia->object_dialogs;
             DIALOG *pointed = NULL;

             // Find the object being pointed to
             pointed = find_mouse_object(current);

             // Make sure that we pointed to an object
             if(pointed==NULL) return D_O_K;

             // Find the index of the pointed to dialog
             int index = -1;
             for(int i=0;current[i].proc;i++)
                {
                if(pointed==&(current[i])) index = i;
                }
             // Make sure that the object isn't pushed as far as we want
             if(index<=6) return D_O_K;

             // Tell what we might do
             if(alert("Push object back?",NULL,NULL,"Yes","No",'y','n')==2) return D_O_K;

             // Get the two objects that will be swapped
             dialog_object *front = (dialog_object*)current[index-1].dp;
             dialog_object *back = (dialog_object*)current[index].dp;

             // Swap the dialogs and connect properly to kkep sizes a flags proper
             DIALOG temp = current[index-1];
             front->connect(&temp);

             current[index-1] = current[index];
             back->connect(&current[index-1]);

             current[index] = temp;
             front->connect(&current[index]);

             ret = D_REDRAW | D_USED_CHAR;
             }
          // Check for an object pullfront
          else if((c>>8)==dia->pullfront_key)
             {
             // Get the Allegro dialogs
             DIALOG *current = dia->object_dialogs;
             DIALOG *pointed = NULL;

             // Find the object being pointed to
             pointed = find_mouse_object(current);

             // Make sure that we pointed to an object
             if(pointed==NULL) return D_O_K;

             // Find the index of the pointed to dialog
             int index = -1;
             for(int i=0;current[i].proc;i++)
                {
                if(pointed==&(current[i])) index = i;
                }
             // Make sure it was really in there
             if(current[index].proc==NULL) return D_O_K;

             // Make sure it was not pulled as far as possible allready
             if(current[index+1].proc==NULL) return D_O_K;

             // Make sure that the object isn't pulled as far as we want
             if(index>=dia->number) return D_O_K;

             // Tell what we might do
             if(alert("Pull object to front?",NULL,NULL,"Yes","No",'y','n')==2) return D_O_K;

             // Get the two objects that will be swapped
             dialog_object *front = (dialog_object*)current[index+1].dp;
             dialog_object *back = (dialog_object*)current[index].dp;

             // Swap the dialogs and connect properly to kkep sizes a flags proper
             DIALOG temp = current[index+1];
             front->connect(&temp);

             current[index+1] = current[index];
             back->connect(&current[index+1]);

             current[index] = temp;
             front->connect(&current[index]);

             ret = D_REDRAW | D_USED_CHAR;
             }
          // Check for an object forground color change
          else if((c>>8)==dia->fg_color_key)
             {
             // Find the object being pointed to
             DIALOG *current = dia->object_dialogs;
             DIALOG *pointed = NULL;
             pointed = find_mouse_object(current);

             // Adjust th object color if there is an object
             if(pointed!=NULL)
                {
                adjust_object_color(&(pointed->fg),"Modify object foreground color");
                dia->redraw();
                }
             }
          // Check for an object background color change
          else if((c>>8)==dia->bg_color_key)
             {
             // Find the object being pointed to
             DIALOG *current = dia->object_dialogs;
             DIALOG *pointed = NULL;
             pointed = find_mouse_object(current);

             // Adjust th object color if there is an object
             if(pointed!=NULL)
                {
                adjust_object_color(&(pointed->bg),"Modify object background color");
                dia->redraw();
                }
             }
          else
             {
             // Recieved an unknown MSG_XCAR
             }
          break;
          }
       default:
          break;
       }
    return ret;
    }

//----------------------------------------------------------------------------
// Load the dialogs from disk
bool builder_dialog::load (const char *filename, const bool append)
	{
   // Try to open the file
   FILE *stream = fopen(filename,"rt");
   if(stream==NULL) return false;

   // Load the header and see if it's OK
   char temp[2000];
   fgets(temp,999,stream);
   if(strcmp(temp,dialog_file_header)==1)
      {
      fclose(stream);
      return false;
      }

   // Remove the existing objects if this is not an append
   if(append==false)
      {
      // Remove allocated objects everything
      for(int i=6;i<number;i++) object_dialogs[i].proc = empty_proc;
      }

   // Read in each line and get it decoded
   while(true)
      {
      // Read in a line for decoding
      fgets(temp,999,stream);
      if(strlen(temp)<3) break;

      // Decode the object
      object_description *desc = new object_description;
      dialog_object *object = desc->decode(temp);
      object->dataptr = desc;

      // Now get more info about placement and colour
      int x, y, w, h;
      int fr, fg, fb, br, bg, bb;
      fgets(temp,999,stream);
      sscanf(temp,"%i,%i,%i,%i,%i,%i,%i,%i,%i,%i",&x,&y,&w,&h,&fr,&fg,&fb,&br,&bg,&bb);

      // Calculate the correct color from the 0-255 RGB values
      int fore = makecol8(fr,fg,fb);
      int back = makecol8(br,bg,bb);

      // Modify the object stuff
      object->set_x(x);
      object->set_y(y);
      object->set_w(w);
      object->set_h(h);
      object->set_fg(fore);
      object->set_bg(back);

      // Now add the object to our builder dialog
      add(object);
      }
   fclose(stream);
   return true;
	}
//----------------------------------------------------------------------------
// Save the dialogs to disk
void builder_dialog::save (const char *filename)
	{
   // Try to delete the file and open a new one
   FILE *stream = fopen(filename,"wt");
   if(stream==NULL) return;

   fprintf(stream,"%s",dialog_file_header);

   // Colour used for output
   RGB fore;
   RGB back;

   // Get the right list
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   // Write the object declarations out
   char temp[2000];
   for(int i=6;i<number;i++)
      {
      // Find the description
      dialog_object *object = (dialog_object*)list[i].dp;
      object_description *descript = (object_description*)object->dataptr;

      // Encode the objects
      descript->encode(temp);

      // Send the encoded stuff to file
      fprintf(stream,"%s\n",temp);

      // Send out the placements
      fprintf(stream,"%i,%i,%i,%i,",object_dialogs[i].x,object_dialogs[i].y,object_dialogs[i].w,object_dialogs[i].h);

      // Send out the colors in RGB
      get_color(list[i].fg,&fore);
      get_color(list[i].bg,&back);
      fprintf(stream,"%i,%i,%i,%i,%i,%i\n",fore.r*4,fore.g*4,fore.b*4,back.r*4,back.g*4,back.b*4);
      }
   fprintf(stream,"\n\n");
   fclose(stream);
	}
//----------------------------------------------------------------------------
// Export a C++ object file
void builder_dialog::export (const char *filename)
   {
   // Try to open the file
   FILE *stream = fopen(filename,"wt");
   if(stream==NULL) return;

   fprintf(stream,"   // Object definitions - DEGUI %s\n",DEGUI_VERSION_STR);

   // Get the right list
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   // Write the object declarations out
   for(int i=6;i<number;i++)
      {
      // Find the description
      dialog_object *object = (dialog_object*)list[i].dp;
      object_description *descript = (object_description*)object->dataptr;

      // Check if there are any arguments
      const char *args = descript->get_arguments();

      // Output for no arguments
      if(strcmp(args,"NULL")==0)
         {
         fprintf(stream,"   %s %s;\n",descript->get_type(),descript->get_name());
         }
      // Output for special arguments
      else
         {
         fprintf(stream,"   %s %s(%s);\n",descript->get_type(),descript->get_name(),args);
         }
      }

   fprintf(stream,"\n   // Add Objects to the dialog\n");

   // Make a dialog and add it to them
   fprintf(stream,"   dialog the_dialog;\n");
   for(int i=6;i<number;i++)
      {
      // Find the description
      dialog_object *object = (dialog_object*)list[i].dp;
      object_description *descript = (object_description*)object->dataptr;

      fprintf(stream,"   the_dialog.add(%s,",descript->get_name());
      fprintf(stream,"%i,%i,%i,%i,%i,%i);\n",object->x(),object->y(),object->w(),object->h(),object->fg(),object->bg());
      }
   // Possibly center the dialog
   if(alert("Should dialog be centered?",NULL,NULL,"&Yes","&No",'y','n')==1)
      {
      fprintf(stream,"   the_dialog.center();");
      }
   // Possibly popup or execute the dialog
   if(alert("Should dialog replace it's","background when done?",NULL,"&Yes","&No",'y','n')==1)
      {
      fprintf(stream,"   the_dialog.popup();\n");
      }
   else
      {
      fprintf(stream,"   the_dialog.execute();\n\n\n");
      }
   fclose(stream);
   }

//----------------------------------------------------------------------------
// Object adding procedures
//----------------------------------------------------------------------------
int builder_dialog::add_box_object (void)
   {
   // Create the object
   box_object *object = new box_object;
   object_description *desc = new object_description;
	if((object==NULL)||(desc==NULL)) degui_no_memory();

   // Describe the object
   desc->set_type(box_type);

   // Link the object and the description
   object->dataptr = desc;

   // Add the object to the dialog
   add(object);
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_shadow_box_object (void)
   {
   // Create the object
   shadow_box_object *object = new shadow_box_object;
   object_description *desc = new object_description;
	if((object==NULL)||(desc==NULL)) degui_no_memory();

   // Describe the object
   desc->set_type(shadow_box_type);

   // Link the object and the description
   object->dataptr = desc;

   add(object);
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_bitmap_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char bit_name[50];
   strcpy(bit_name,"bitmap");

   char filename[80];
   strcpy(filename,"bitmap");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_4("Variable name:");
   edittext_object object_9(bit_name,49);
   edittext_object object_11(variable_name,49);
   text_object bitmap_name("Bitmap name:");
   text_object object_7("Create bitmap object");
   checkbox_object loadable("Load bitmap from file");
   checkbox_object use_pal("Use palette from file");

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_4,176,256,110,10);
   the_dialog.add(object_9,272,200,192,13,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,288,256,176,20,gui_fg_color,gui_mg_color);
   the_dialog.add(bitmap_name,176,200,96,8);
   the_dialog.add(object_7,239,170,160,8);
   the_dialog.add(loadable,176,216,280,8);
   the_dialog.add(use_pal,176,232,268,8);
   the_dialog.center();

   // Execute the dialog and go further if it's OK'd
   if(the_dialog.popup()==&object_1)
      {
      // Check if we load the bitmap from a file
      if((loadable.selected()==true)
         &&(file_select("Load bitmap object from file",filename,"bmp;tga;lbm;pcx")!=0))
         {
         // Create the object
         helper_bitmap_object *object = new helper_bitmap_object(filename);
         object_description *desc = new object_description;
         if((object==NULL)||(desc==NULL)) degui_no_memory();

         // Set the variable name if the user entered one
         desc->set_type(bitmap_type);
         object->dataptr = desc;

         char temp[100];
         sprintf(temp,"\"%s\"",filename);

         // Exchange the slashes with backslashes or we will have problems
         // that the slashes look like escape sequences in the char* in
         // the constructor.  uuuggghhh... wierd...
         char *t;
         while((t=strchr(temp,'\\'))!=NULL)
            {
            t[0] = '/';
            t++;
            }
         desc->set_arguments(temp);

         if(strlen(variable_name)>1) desc->set_name(variable_name);

         add(object);

         // Possibly modify the system palette
         if(use_pal.selected()==true)
            {
            PALETTE pal;
            BITMAP *b;

            // Find the type of file it is and load it
            if(stricmp(get_extension((char*)filename),"bmp")==0) b = load_bmp((char*)filename,pal);
            else if(stricmp(get_extension((char*)filename),"tga")==0) b = load_tga((char*)filename,pal);
            else if(stricmp(get_extension((char*)filename),"lbm")==0) b = load_lbm((char*)filename,pal);
            else b = load_pcx((char*)filename,pal);

            // Tell the system to use this palette
            use_palette(pal);
            destroy_bitmap(b);
            }
         }
      // Or just use an empty bitmap
      else
         {
         // Create the object
         helper_bitmap_object *object = new helper_bitmap_object(bit_name);
         object_description *desc = new object_description;
         if((object==NULL)||(desc==NULL)) degui_no_memory();

         // Set the variable name if the user entered one
         desc->set_type(bitmap_type);
         object->dataptr = desc;

         desc->set_arguments(bit_name);

         if(strlen(variable_name)>1) desc->set_name(variable_name);

         add(object);
         }
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_text_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char initial_text[50];
   char pointer_name[50];
   strcpy(initial_text,"Text");
   strcpy(pointer_name,"pointer");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_3("Create text object");
   text_object object_4("Variable name:");
   radio_button_object object_7("Object text:",0,false);
   radio_button_object object_8("Pointer text:",0,false);
   edittext_object object_9(initial_text,49);
   edittext_object object_10(pointer_name,49);
   edittext_object object_11(variable_name,49);

   object_7.select();  // Use the normal text as default

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_3,240,170,140,10);
   the_dialog.add(object_4,175,255,110,10);
   the_dialog.add(object_7,175,205,100,8);
   the_dialog.add(object_8,175,230,100,8);
   the_dialog.add(object_9,290,205,155,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_10,295,230,150,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,290,255,175,10,gui_fg_color,gui_mg_color);
   the_dialog.center();

   // Get some data from the user
   helper_text_object *object;
   if(the_dialog.popup()==&object_1)
      {
      char temp[100];

      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;
      desc->set_type(text_type);

      // Create the object
      if(object_7.selected()==true)
         {
         object = new helper_text_object(initial_text);
         sprintf(temp,"\"%s\"",initial_text);
         }
      else
         {
         object = new helper_text_object(pointer_name);
         sprintf(temp,"%s",pointer_name);
         }
		if(object==NULL) return D_REDRAW;

      desc->set_arguments(temp);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_centerd_text_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char initial_text[50];
   char pointer_name[50];
   strcpy(initial_text,"Text");
   strcpy(pointer_name,"pointer");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_3("Create centerd text object");
   text_object object_4("Variable name:");
   radio_button_object object_7("Object text:",0,false);
   radio_button_object object_8("Pointer text:",0,false);
   edittext_object object_9(initial_text,49);
   edittext_object object_10(pointer_name,49);
   edittext_object object_11(variable_name,49);

   object_7.select();  // Use the normal text as default

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_3,220,170,140,10);
   the_dialog.add(object_4,175,255,110,10);
   the_dialog.add(object_7,175,205,100,8);
   the_dialog.add(object_8,175,230,100,8);
   the_dialog.add(object_9,290,205,155,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_10,295,230,150,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,290,255,175,10,gui_fg_color,gui_mg_color);
   the_dialog.center();

   // Get some data from the user
   helper_centerd_text_object *object;
   if(the_dialog.popup()==&object_1)
      {
      char temp[100];

      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;
      desc->set_type(centerd_text_type);

      // Create the object
      if(object_7.selected()==true)
         {
         object = new helper_centerd_text_object(initial_text);
         sprintf(temp,"\"%s\"",initial_text);
         }
      else
         {
         object = new helper_centerd_text_object(pointer_name);
         sprintf(temp,"%s",pointer_name);
         }
		if(object==NULL) return D_REDRAW;

      desc->set_arguments(temp);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   reset();
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_edittext_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char pointer_name[50];
   strcpy(pointer_name,"pointer");

   char length_string[5];
   strcpy(length_string,"10");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_4("Variable name:");
   edittext_object object_9(pointer_name,14);
   edittext_object object_10(length_string,4);
   edittext_object object_11(variable_name,49);
   text_object object_7("Create editable text object");
   text_object object_8("Pointer name:");
   text_object object_6("Max length:");

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_4,175,255,110,10);
   the_dialog.add(object_9,290,205,155,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_10,290,230,155,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,285,255,175,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_7,210,170,215,10);
   the_dialog.add(object_8,185,205,100,8);
   the_dialog.add(object_6,200,230,85,10);
   the_dialog.center();

   // Set a minimum maximum length
   int len = atoi(length_string);
   if(len<1) len = 1;

   // Get some data from the user
   edittext_object *object;
   if(the_dialog.popup()==&object_1)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Create the object
      object = new edittext_object(NULL,len);
		if(object==NULL) return D_REDRAW;

      // Setup the description
      desc->set_type(edittext_type);

      char temp[100];
      sprintf(temp,"%s,%i",pointer_name,len);
      desc->set_arguments(temp);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_button_object (void)
   {
   char name[80];
   strcpy(name,"Button");

   // Object definitions
   shadow_box_object object_0;
   text_object object_1("Button Object Properties");
   text_object object_2("Button text:");
   edittext_object object_3(name,79);
   checkbox_object object_4("Close dialog on click");
   button_object object_5("&Ok",'o',true);
   button_object object_6("&Cancel",'c',true);

   // Add objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,100,100,260,135);
   the_dialog.add(object_1,135,110,190,10);
   the_dialog.add(object_2,115,135,100,8);
   the_dialog.add(object_3,115,145,200,8,gui_fg_color,gui_mg_color);
   the_dialog.add(object_4,115,165,185,10);
   the_dialog.add(object_5,110,195,115,30);
   the_dialog.add(object_6,240,195,110,30);
   the_dialog.center();

   if(the_dialog.popup()==&object_5)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Find what the user wants
      bool clos = false;
      if(object_4.selected()==true) clos = true;

      helper_button_object *object = new helper_button_object(name,0,clos);
		if(object==NULL) return D_REDRAW;

      // Setup the description
      desc->set_type(button_type);

      char temp[100];
      sprintf(temp,"\"%s\",0,",name);
      if(clos==true) strcat(temp,"true");
      else strcat(temp,"false");

      desc->set_arguments(temp);

      object->dataptr = desc;

      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_checkbox_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char initial_text[50];
   strcpy(initial_text,"Checkbox");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_3("Create checkbox object");
   text_object object_4("Variable name:");
   text_object object_7("  Object text:");
   edittext_object object_9(initial_text,49);
   edittext_object object_11(variable_name,49);

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_3,240,170,140,10);
   the_dialog.add(object_4,175,245,110,10);
   the_dialog.add(object_7,175,215,100,8);
   the_dialog.add(object_9,290,215,155,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,290,245,175,10,gui_fg_color,gui_mg_color);
   the_dialog.center();

   // Get some data from the user
   if(the_dialog.popup()==&object_1)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Create the object
      helper_checkbox_object *object = new helper_checkbox_object(initial_text);
		if(object==NULL) return D_REDRAW;

      // Setup the description
      desc->set_type(checkbox_type);

      char temp[100];
      sprintf(temp,"\"%s\"",initial_text);
      desc->set_arguments(temp);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_radio_button_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char initial_text[50];
   strcpy(initial_text,"Radio button");

   char group_string[10];
   strcpy(group_string,"0");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_4("Variable name:");
   edittext_object object_9(initial_text,49);
   edittext_object object_10(group_string,4);
   edittext_object object_11(variable_name,49);
   text_object object_text("Object text:");
   text_object object_12("Create radio button object");
   text_object group("Group number:");
   checkbox_object squareness("Square");

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_4,175,255,110,10);
   the_dialog.add(object_9,275,205,190,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_10,280,220,30,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,290,255,175,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_text,175,205,100,8);
   the_dialog.add(object_12,210,170,210,10);
   the_dialog.add(group,175,220,100,8);
   the_dialog.add(squareness,175,235,100,8);
   the_dialog.center();

   // Get some data from the user
   helper_radio_button_object *object;
   if(the_dialog.popup()==&object_1)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Create the object
      object = new helper_radio_button_object(initial_text,atoi(group_string),squareness.selected());
		if(object==NULL) return D_REDRAW;

      // Setup the description
      desc->set_type(radio_button_type);

      char temp[100];
      sprintf(temp,"\"%s\",%i,",initial_text,atoi(group_string));
      if(squareness.selected()==true) strcat(temp,"true");
      else strcat(temp,"false");
      desc->set_arguments(temp);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_list_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char procedure_name[50];
   strcpy(procedure_name,"default");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_4("Variable name:");
   edittext_object object_10(procedure_name,14);
   edittext_object object_11(variable_name,14);
   text_object object_13("List procedure must be of the type");
   text_object object_8("Procedure:");
   text_object object_9("char* (*proc)(int, int*);");
   text_object object_12("Create list object");

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_4,175,255,110,10);
   the_dialog.add(object_10,255,230,210,15,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,290,255,175,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_13,175,200,100,8);
   the_dialog.add(object_8,175,230,80,10);
   the_dialog.add(object_9,175,210,100,8);
   the_dialog.add(object_12,240,170,100,8);
   the_dialog.center();

   // Check if the default procedure has been chosen
   char *proc_name = NULL;
   if(stricmp(procedure_name,"default")!=0) proc_name = procedure_name;

   // Get some data from the user
   list_object *object;
   if(the_dialog.popup()==&object_1)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Create the object
      object = new list_object;
		if(object==NULL) return D_REDRAW;

      // Setup the description
      desc->set_type(list_type);

      desc->set_arguments(proc_name);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_textbox_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char pointer_name[50];
   strcpy(pointer_name,"pointer");

   // Object definitions - DEGUI 1.0 beta
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_4("Variable name:");
   edittext_object object_11(variable_name,49);
   text_object title("Create textbox object");
   text_object object_6("Pointer:");
   edittext_object edit_pname(pointer_name,49);
   checkbox_object use_wrap("Word wrap");
   checkbox_object text_wrap("Text wrap");

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_4,175,260,110,10);
   the_dialog.add(object_11,290,260,170,20);
   the_dialog.add(title,230,170,168,8);
   the_dialog.add(object_6,185,205,64,8);
   the_dialog.add(edit_pname,250,205,210,10);
   the_dialog.add(use_wrap,185,240,88,8);
   the_dialog.add(text_wrap,185,225,88,8);

   // Get some data from the user
   textbox_object *object;
   if(the_dialog.popup()==&object_1)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Create the object
		object = new textbox_object(NULL,text_wrap.selected(),use_wrap.selected());

      // Setup the description
      desc->set_type(textbox_type);

      char temp[100];
      sprintf(temp,"%s,",pointer_name);
      if(text_wrap.selected()==true) strcat(temp,"true,");
      else strcat(temp,"false,");
      if(use_wrap.selected()==true) strcat(temp,"true");
      else strcat(temp,"false");
      desc->set_arguments(temp);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }
//----------------------------------------------------------------------------
int builder_dialog::add_user_object (void)
   {
   // Data definitions
   char variable_name[50];
   strcpy(variable_name,"");

   char type_name[50];
   strcpy(type_name,"user_type");

   char args[100];
   strcpy(args,"");

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_4("Variable name:");
   edittext_object object_6(type_name,49);
   edittext_object object_11(variable_name,49);
   text_object object_8("Type:");
   text_object object_12("Arguments:");
   edittext_object object_14(args,99);
   text_object object_13("Create user object");

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,165,160,305,160);
   the_dialog.add(object_1,215,285,80,20);
   the_dialog.add(object_2,350,285,80,20);
   the_dialog.add(object_4,175,260,110,10);
   the_dialog.add(object_6,215,200,240,15,gui_fg_color,gui_mg_color);
   the_dialog.add(object_11,290,260,175,20,gui_fg_color,gui_mg_color);
   the_dialog.add(object_8,175,200,40,10,gui_fg_color,gui_mg_color);
   the_dialog.add(object_12,175,230,100,8);
   the_dialog.add(object_14,175,240,290,15);
   the_dialog.add(object_13,245,170,100,8);
   the_dialog.center();

   // Get some data from the user
   helper_object *object;
   if(the_dialog.popup()==&object_1)
      {
      // Allocate the object description
      object_description *desc = new object_description;
      if(desc==NULL) return D_REDRAW;

      // Create the text for the user object
      char temp[100];
      sprintf(temp,"%s",type_name);

      if(strlen(args)>1)
         {
         strcat(temp,"(");
         strcat(temp,args);
         strcat(temp,")");
         }
      strcat(temp,";");

      // Create the object
      object = new helper_object(temp);
		if(object==NULL) return D_REDRAW;

      // Setup the description
      desc->set_type(type_name);

      desc->set_arguments(args);

      object->dataptr = desc;

      // Set the variable name if the user entered one
      if(strlen(variable_name)>1) desc->set_name(variable_name);

      // Add the object to the gui
      add(object);
      }
   return D_REDRAW;
   }

//----------------------------------------------------------------------------
// Helper function to adjust an objects color
void adjust_object_color (int *color, const char *title)
   {
   // Set up the default colours and indexes
   char red_string[5];
   char green_string[5];
   char blue_string[5];
   char index_string[5];

   RGB ground;
   get_color(*color,&ground);

   sprintf(index_string,"%i",*color);
   sprintf(red_string,"%i",ground.r*4);
   sprintf(green_string,"%i",ground.g*4);
   sprintf(blue_string,"%i",ground.b*4);

   // Object definitions
   shadow_box_object object_0;
   button_object object_1("&Ok",0,true);
   button_object object_2("&Cancel",0,true);
   text_object object_3("Red:");
   text_object object_4("Green:");
   text_object object_5("Blue:");
   text_object object_6("Index:");
   edittext_object object_7(red_string,4);
   edittext_object object_8(green_string,4);
   edittext_object object_9(blue_string,4);
   edittext_object object_10(index_string,4);
   radio_button_object use_pal("Use palette index",0,false);
   checkbox_object force("Force nearest color to these RGB");
   radio_button_object use_RGB("Use RGB values (0-255)",0,false);
   text_object object_14(title);

   use_RGB.select();

   // Add Objects to the dialog
   dialog the_dialog;
   the_dialog.add(object_0,180,179,300,168);
   the_dialog.add(object_1,226,307,80,20);
   the_dialog.add(object_2,354,307,80,20);
   the_dialog.add(object_3,198,231,32,8);
   the_dialog.add(object_4,270,231,48,8);
   the_dialog.add(object_5,358,231,40,8);
   the_dialog.add(object_6,198,283,48,8);
   the_dialog.add(object_7,230,231,36,8,gui_fg_color,gui_mg_color);
   the_dialog.add(object_8,318,231,36,8,gui_fg_color,gui_mg_color);
   the_dialog.add(object_9,398,231,36,8,gui_fg_color,gui_mg_color);
   the_dialog.add(object_10,246,283,36,8,gui_fg_color,gui_mg_color);
   the_dialog.add(use_pal,198,271,152,8);
   the_dialog.add(force,198,243,284,8);
   the_dialog.add(use_RGB,198,219,128,8);
   the_dialog.add(object_14,202,191,240,8);
   the_dialog.center();

   // Execute the dialog and see if it was OK'd
   if(the_dialog.popup()==&object_1)
      {
      // Check if we use RGB vales
      if(use_RGB.selected()==true)
         {
         // Find the index of the closest color
         int red = atoi(red_string);
         int green = atoi(green_string);
         int blue = atoi(blue_string);

         *color = makecol8(red,green,blue);

         // Check if we force the closest color to be exactly that RGB values
         if(force.selected()==true)
            {
            RGB actual;
            actual.r = red/4;
            actual.g = green/4;
            actual.b = blue/4;
            set_color(*color,&actual);
            }
         }
      // Or do we use palette index
      else
         {
         *color = atoi(index_string);
         }
      }
   }
//----------------------------------------------------------------------------

