//----------------------------------------------------------------------------
//
// C++ Objects for Allegro's gui
//
// Douglas Eleveld (D.J.Eleveld@anest.azg.nl)
//
//----------------------------------------------------------------------------
#include "degui.h"
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

//----------------------------------------------------------------------------
// Stuff for handling the default behaviour in callback functions
int dialog_object::default_message = 0;
dialog_object* dialog_object::default_object = NULL;
int dialog_object::default_data = 0;
int dialog_object::default_answer = D_O_K;

//----------------------------------------------------------------------------
// The main dispatching function i.e. the interface between the C and C++ object
int dialog_object::object_message_proc (int msg, DIALOG *d, int c)
   {
   // Make sure that there IS an object there
   if(d->dp==NULL) return D_O_K;

   // Possibly problematic cast
   dialog_object *object = (dialog_object*)(d->dp);

   // Tell the object what it's dialog is
   if(object->d!=d) object->connect(d);

   // Possibly do a callback on every message
   if(object->all_callback!=NULL) object->all_callback();

   // Setup any default behaviour
   dialog_object::default_message = msg;
   dialog_object::default_object = object;
   dialog_object::default_data = c;
   dialog_object::default_answer = D_O_K;

   bool request_close = false;
   int callback_ret = D_O_K;

   // Convert the allegro message to a call of a member function
   // or the call of a callback function
   switch(msg)
      {
      case(MSG_START):
         if(object->start_callback!=NULL) callback_ret = object->start_callback();
         else object->msg_start();
         break;
      case(MSG_END):
         if(object->end_callback!=NULL) callback_ret = object->end_callback();
         else object->msg_end();
         break;
      case(MSG_DRAW):
         if(object->draw_callback!=NULL) callback_ret = object->draw_callback();
         else object->msg_draw();
         object->_redraw = false;
         break;
      case(MSG_CLICK):
         if(object->click_callback!=NULL) callback_ret = object->click_callback();
         else object->msg_click();

			// Check if we exit on click
			if(object->d->flags&D_EXIT) request_close = true;
         break;
      case(MSG_DCLICK):
         if(object->dclick_callback!=NULL) callback_ret = object->dclick_callback();
         else object->msg_dclick();
         break;
      case(MSG_KEY):
         if(object->key_callback!=NULL) callback_ret = object->key_callback();
         else object->msg_key();

			// Check if we exit on keypress
			if(object->d->flags&D_EXIT) request_close = true;
         break;
      case(MSG_CHAR):
         if(object->char_callback!=NULL) callback_ret = object->char_callback(c);
         else if(object->msg_char(c)==true) return D_USED_CHAR;
      case(MSG_XCHAR):
         if(object->xchar_callback!=NULL) callback_ret = object->xchar_callback(c);
         else if(object->msg_xchar(c)==true) return D_USED_CHAR;
      case(MSG_WANTFOCUS):
         if(object->wantfocus_callback!=NULL) callback_ret = object->wantfocus_callback();
         else if(object->msg_wantfocus()==true) return D_WANTFOCUS;
      case(MSG_GOTFOCUS):
         if(object->gotfocus_callback!=NULL) callback_ret = object->gotfocus_callback();
         else object->msg_gotfocus();
         break;
      case(MSG_LOSTFOCUS):
         if(object->lostfocus_callback!=NULL) callback_ret = object->lostfocus_callback();
         else object->msg_lostfocus();
         break;
      case(MSG_GOTMOUSE):
         if(object->gotmouse_callback!=NULL) callback_ret = object->gotmouse_callback();
         else object->msg_gotmouse();
         break;
      case(MSG_LOSTMOUSE):
         if(object->lostmouse_callback!=NULL) callback_ret = object->lostmouse_callback();
         else object->msg_lostmouse();
         break;
      case(MSG_IDLE):
         if(object->idle_callback!=NULL) callback_ret = object->idle_callback();
         else object->msg_idle();

         // Check if the object needs a redraw during the idle period
         if(object->_redraw==true)
            {
            if(object->draw_callback!=NULL) callback_ret = object->draw_callback();
            else object->msg_draw();
            object->_redraw = false;
            }
         break;
      case(MSG_RADIO):
         if(object->radio_callback!=NULL) callback_ret = object->radio_callback(c);
         else object->msg_radio(c);
         break;
      default:
         if(object->unknown_callback!=NULL) callback_ret = object->unknown_callback(msg);
         else object->msg_unknown(msg);
         break;
      }

   // Make sure that any wierd calls to the default behaviour don't do anything
   dialog_object::default_message = MSG_IDLE;
   dialog_object::default_object = NULL;
   dialog_object::default_data = 0;

   // Construct the Allegro style answer
   // If we called a callback function then the dialog_object::default_answer
   // will be set correctly else it will just be D_O_K like we set it at the
   // beginning of this function
   int ret = dialog_object::default_answer;

   // Modify the return value depending on what the
   // callback funtion has asked for
   ret |= callback_ret;

   if(request_close==true) ret |= D_CLOSE;

   return ret;
   }

//----------------------------------------------------------------------------
// Dialog base class object functions
//----------------------------------------------------------------------------
// Constructor
dialog_object::dialog_object (void)
   :d(new DIALOG),
   connected(false),
   in_c_dialog(false),
   _parent(NULL),
   _redraw(false),
   all_callback(NULL),
   start_callback(NULL),
   end_callback(NULL),
   draw_callback(NULL),
   click_callback(NULL),
   dclick_callback(NULL),
   key_callback(NULL),
   char_callback(NULL),
   xchar_callback(NULL),
   wantfocus_callback(NULL),
   gotfocus_callback(NULL),
   lostfocus_callback(NULL),
   gotmouse_callback(NULL),
   lostmouse_callback(NULL),
   idle_callback(NULL),
   radio_callback(NULL),
   unknown_callback(NULL),
   dataptr(NULL)
   {
   // Check the d pointer
   if(d==NULL) degui_no_memory();

   d->proc = object_message_proc;
   d->x = 0;
   d->y = 0;
   d->w = 0;
   d->h = 0;
   d->fg = gui_fg_color;
   d->bg = gui_bg_color;
   d->key = 0;
   d->flags = 0;
   d->d1 = 0;
   d->d2 = 0;
   d->dp = this;
   }
//----------------------------------------------------------------------------
// Destructor
dialog_object::~dialog_object (void)
   {
   if(connected==false) delete d;
   }
//----------------------------------------------------------------------------
// Basic message passing functions

// Initialization of the dialog
void dialog_object::msg_start (void)      { }

// De-initialization of the dialog
void dialog_object::msg_end (void)        { }

// Tell the object to draw itself
void dialog_object::msg_draw (void)		   { }

// Tell the object to deal with a mouse click default is the same as a keypress
void dialog_object::msg_click (void)      { msg_key(); }

// Tell the object to deal with a double click
void dialog_object::msg_dclick (void)     { }

// Tell the object that a shortcut key was pressed
void dialog_object::msg_key (void)        { }

// Tell the object that a key was pressed while it had the focus
// This should return true if the key was used
bool dialog_object::msg_char (const int)  { return false; }

// Tell the object that a key was pressed while it did noty have the focus
// This should return true if the key was used
bool dialog_object::msg_xchar (const int) { return false; }

// Ask the object if they want the focus
bool dialog_object::msg_wantfocus (void)  { return false; }

// Tell the object that the got or lost the focus
void dialog_object::msg_gotfocus (void)   { }
void dialog_object::msg_lostfocus (void)  { }

// Tell the object that they got or lost the mouse
void dialog_object::msg_gotmouse (void)   { }
void dialog_object::msg_lostmouse (void)  { }

// Tell the object that the object manager is bored
void dialog_object::msg_idle (void)       { }

// Deselect grouped radio buttons
void dialog_object::msg_radio (const int)	{ }

// Get an hitherto unknown message
void dialog_object::msg_unknown (const int) { }

//-----------------------------------------------------------------------------
// Create and use a DIALOG strucure for an object
DIALOG dialog_object::create_DIALOG (const int x, const int y, const int w, const int h, const int fg, const int bg)
   {
   // Place the object
   if(x>=0) d->x = x;
   if(y>=0) d->y = y;
   if(w>=0) d->w = w;
   if(h>=0) d->h = h;
   if(fg>=0) d->fg = fg;
   if(bg>=0) d->bg = bg;

   in_c_dialog = true;

   return *d;
   }
//----------------------------------------------------------------------------
// Connect and disconnect the dialog_object from a DIALOG
// When not connected, it has it's own allocated DIALOG
void dialog_object::connect (DIALOG *newd)
   {
   // Make sure we don't connect to ourselves
   if(d==newd) return;

   if(in_c_dialog==false)
      {
      newd->x = d->x;
      newd->y = d->y;
      newd->w = d->w;
      newd->h = d->h;
      newd->key = d->key;
      newd->flags = d->flags;
      newd->d1 = d->d1;
      newd->d2 = d->d2;
      }

   // Set the new d pointer
   if(connected==false)
      {
      delete d;
      connected = true;
      in_c_dialog = false;
      }
   d = newd;
   }
//----------------------------------------------------------------------------
void dialog_object::disconnect (void)
   {
   if(connected==true)
      {
      // Make the new d pointer
      DIALOG *newd = new DIALOG;

      // Copy init info over
      newd->x = d->x;
      newd->y = d->y;
      newd->w = d->w;
      newd->h = d->h;
      newd->key = d->key;
      newd->flags = d->flags;
      newd->d1 = d->d1;
      newd->d2 = d->d2;

      // Set the new d pointer
      d = newd;
      connected = false;
      in_c_dialog = false;
      _parent = NULL;
      }
   }

//----------------------------------------------------------------------------
// The default behaviour in callback functions
//----------------------------------------------------------------------------
void dialog_object::default_behaviour (void)
   {
   // Make sure we have been called properly
   if(default_object==NULL) return;

	// Convert the allegro message to a call of a member function
   switch(default_message)
      {
      case(MSG_START):
         default_object->msg_start();
         break;
      case(MSG_END):
         default_object->msg_end();
         break;
      case(MSG_DRAW):
         default_object->msg_draw();
         break;
      case(MSG_CLICK):
         default_object->msg_click();

			// Check if we exit on keypress
			if(default_object->d->flags&D_EXIT) dialog_object::default_answer |= D_CLOSE;
         break;
      case(MSG_DCLICK):
         default_object->msg_dclick();
         break;
      case(MSG_KEY):
         default_object->msg_key();

			// Check if we exit on keypress
			if(default_object->d->flags&D_EXIT) dialog_object::default_answer |= D_CLOSE;
         break;
      case(MSG_CHAR):
         if(default_object->msg_char(default_data)==true) dialog_object::default_answer = D_USED_CHAR;
         break;
      case(MSG_XCHAR):
         if(default_object->msg_xchar(default_data)==true) dialog_object::default_answer = D_USED_CHAR;
         break;
      case(MSG_WANTFOCUS):
         if(default_object->msg_wantfocus()==true) dialog_object::default_answer = D_WANTFOCUS;
         break;
      case(MSG_GOTFOCUS):
         default_object->msg_gotfocus();
         break;
      case(MSG_LOSTFOCUS):
         default_object->msg_lostfocus();
         break;
      case(MSG_GOTMOUSE):
			default_object->msg_gotmouse();
         break;
      case(MSG_LOSTMOUSE):
         default_object->msg_lostmouse();
         break;
      case(MSG_IDLE):
         default_object->msg_idle();
         break;
      case(MSG_RADIO):
         default_object->msg_radio(default_data);
         break;
      default:
         default_object->msg_unknown(default_message);
         break;
      }
   }

//----------------------------------------------------------------------------
// Wrapper class for objects based on Allegro procedures
//----------------------------------------------------------------------------
// Constructor
proc_object::proc_object (int(*pr)(int,DIALOG*,int), void **p)
   :dialog_procedure(pr),
   ptr_dp(p)
   {
   }
//----------------------------------------------------------------------------
// Destructor
proc_object::~proc_object (void)
   {
   }
//----------------------------------------------------------------------------
// Send a message to an underlying C procedure
int proc_object::c_message (const int msg, const int c)
   {
   int ret = D_O_K;

   // Save the data pointer and the procedure pointer
   void *temp_dp = d->dp;
   int(*temp_proc)(int,DIALOG*,int) = d->proc;

   // Set what the procedure wants to see into the data pointer and procedure
   if(ptr_dp!=NULL) d->dp = *ptr_dp;
   else d->dp = NULL;
   d->proc = dialog_procedure;

   // Actually call the C procedure
   if(dialog_procedure!=NULL) ret = dialog_procedure(msg,d,c);

   // Restore any pointers changed
   d->dp = temp_dp;
   d->proc = temp_proc;

   return ret;
   }
//----------------------------------------------------------------------------
// Basic message passing functions

// Initialization of the dialog
void proc_object::msg_start (void)
   {
   dialog_object::msg_start();
   c_message(MSG_START,0);
   }
// De-initialization of the dialog
void proc_object::msg_end (void)
   {
   c_message(MSG_END,0);
   }
// Tell the object to draw itself
void proc_object::msg_draw (void)
   {
   c_message(MSG_DRAW,0);
   }
// Tell the object to deal with a mouse click
void proc_object::msg_click (void)
   {
   c_message(MSG_CLICK,0);
   }
// Tell the object to deal with a double click
void proc_object::msg_dclick (void)
   {
   c_message(MSG_DCLICK,0);
   }
// Tell the object that a shortcut key was pressed
void proc_object::msg_key (void)
   {
   c_message(MSG_KEY,0);
   }
// Tell the object that a key was pressed while it had the focus
// This should return true if the key was used
bool proc_object::msg_char (const int c)
   {
   int ret = c_message(MSG_CHAR,c);

   // Find out if the char was used or not
   if(ret&D_USED_CHAR) return true;
   else return false;
   }
// Tell the object that a key was pressed while it did noty have the focus
// This should return true if the key was used
bool proc_object::msg_xchar (const int c)
   {
   int ret = c_message(MSG_XCHAR,c);

   // Find out if the char was used or not
   if(ret&D_USED_CHAR) return true;
   else return false;
   }
// Ask the object if they want the focus
bool proc_object::msg_wantfocus (void)
   {
   int ret = c_message(MSG_WANTFOCUS,0);

   // Find out if they want the focus or not
   if(ret&D_WANTFOCUS) return true;
   else return false;
   }
// Tell the object that the got or lost the focus
void proc_object::msg_gotfocus (void)
   {
   c_message(MSG_GOTFOCUS,0);
   }
void proc_object::msg_lostfocus (void)
   {
   c_message(MSG_LOSTFOCUS,0);
   }
// Tell the object that they got or lost the mouse
void proc_object::msg_gotmouse (void)
   {
   c_message(MSG_GOTMOUSE,0);
   }
void proc_object::msg_lostmouse (void)
   {
   c_message(MSG_LOSTMOUSE,0);
   }
// Tell the object that the object manager is bored
void proc_object::msg_idle (void)
   {
   c_message(MSG_IDLE,0);
   }
// Deselect grouped radio buttons
void proc_object::msg_radio (const int c)
   {
   c_message(MSG_RADIO,c);
   }
// Get an hitherto unknown message
void proc_object::msg_unknown (const int msg)
   {
   c_message(msg,0);
   }

//----------------------------------------------------------------------------
// Simple object to clear the screen
//----------------------------------------------------------------------------
// Constructor
clear_screen_object::clear_screen_object (void)
   :proc_object(d_clear_proc,NULL)
   {
   // Setup the defaults
   set_x(0);
   set_y(0);
	set_w(0);
   set_h(0);
   }
//----------------------------------------------------------------------------
// Destructor
clear_screen_object::~clear_screen_object (void)
   {
   }

//----------------------------------------------------------------------------
// Simple object just draws a box
//----------------------------------------------------------------------------
// Constructor
box_object::box_object (void)
   :proc_object(d_box_proc,NULL)
   {
   // Setup the defaults
   set_x(100);
   set_y(100);
	set_w(100);
   set_h(100);
   }
//----------------------------------------------------------------------------
// Destructor
box_object::~box_object (void)
   {
   }

//----------------------------------------------------------------------------
// Simple object to draw a shadow box
//----------------------------------------------------------------------------
// Constructor
shadow_box_object::shadow_box_object (void)
   :proc_object(d_shadow_box_proc,NULL)
   {
   // Setup the defaults
   set_x(100);
   set_y(100);
	set_w(100);
   set_h(100);
   }
//----------------------------------------------------------------------------
// Destructor
shadow_box_object::~shadow_box_object (void)
   {
   }

//----------------------------------------------------------------------------
// Simple object draws a bitmap
//----------------------------------------------------------------------------
// Constructor
bitmap_object::bitmap_object (BITMAP* bit)
   :proc_object(d_bitmap_proc,(void**)(&b)),
	b(bit),
   allocated(false)
   {
   // Setup the defaults
   set_x(100);
   set_y(100);
	set_w(100);
   set_h(100);
   }
bitmap_object::bitmap_object (const char *filename)
   :proc_object(d_bitmap_proc,(void**)(&b)),
	b(NULL),
   allocated(false)
   {
   // Setup the defaults
   set_x(100);
   set_y(100);
	set_w(100);
   set_h(100);

   setup(d_bitmap_proc,(void**)(&b));

   // Find the type of file it is and load it
  	PALETTE pal;
   if(stricmp(get_extension((char*)filename),"bmp")==0) b = load_bmp((char*)filename,pal);
   else if(stricmp(get_extension((char*)filename),"tga")==0) b = load_tga((char*)filename,pal);
   else if(stricmp(get_extension((char*)filename),"lbm")==0) b = load_lbm((char*)filename,pal);
   else b = load_pcx((char*)filename,pal);

   // Remember if we have allocated the bitmap or not so we can delete it
   if(b!=NULL)
      {
      allocated = true;

      // The size of the bitmap
      set_w(b->w);
      set_h(b->h);
      }
   }
//----------------------------------------------------------------------------
// Destructor
bitmap_object::~bitmap_object (void)
   {
   if(allocated==true) destroy_bitmap(b);
   }
//----------------------------------------------------------------------------
// Basic message passing functions

// Draw the bitmap maybe
void bitmap_object::msg_draw (void)
	{
   if(b!=NULL) proc_object::msg_draw();
	}

//----------------------------------------------------------------------------
// Simple object draws some text
//----------------------------------------------------------------------------
// Constructor
text_object::text_object (const char* tex)
   :proc_object(d_text_proc,(void**)(&text)),
   text(tex)
   {
   // Setup the defaults
   set_x(100);
   set_y(100);
   set_w(text_length(font,(unsigned char*)text));
   set_h(text_height(font));
   }
//----------------------------------------------------------------------------
// Destructor
text_object::~text_object (void)
   {
   }

//----------------------------------------------------------------------------
// Simple object to draw some ceneterd text
//----------------------------------------------------------------------------
// Constructor
centerd_text_object::centerd_text_object (const char* tex)
   :text_object(tex)
   {
   setup(d_ctext_proc,(void**)(&text));

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_w(text_length(font,(unsigned char*)text)/2);
   set_h(text_height(font));
   }
//----------------------------------------------------------------------------
// Destructor
centerd_text_object::~centerd_text_object (void)
   {
   }

//----------------------------------------------------------------------------
// Button object
//----------------------------------------------------------------------------
// Constructor
button_object::button_object (const char* tex, const char sk, const bool exits)
   :text_object(tex),
	willexit(exits),
	shortcut(sk)
   {
   setup(d_button_proc,(void**)(&text));

	// Setup the exit conditions
	if(willexit) make_exit();
	else dont_exit();

	// Setup the quick key
	set_key(shortcut);

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_w(text_length(font,(unsigned char*)text)+6);
   set_h(text_height(font)+6);
   }
//----------------------------------------------------------------------------
// Destructor
button_object::~button_object (void)
   {
   }

//----------------------------------------------------------------------------
// Simple check box object
//----------------------------------------------------------------------------
// Constructor
checkbox_object::checkbox_object (const char *tex)
   :button_object(tex)
   {
   setup(d_check_proc,(void**)(&text));

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_h(text_height(font));
   set_w(text_length(font,(unsigned char*)text) + 2*h());
   }
//----------------------------------------------------------------------------
// Destructor
checkbox_object::~checkbox_object (void)
   {
   }

//----------------------------------------------------------------------------
// Radio button object
//----------------------------------------------------------------------------
// Constructor
radio_button_object::radio_button_object (const char *tex, const int grou, const bool squa)
   :button_object(tex),
   group(grou),
   square(squa)
   {
   setup(d_radio_proc,(void**)(&text));

   if(square==true) set_d2(1);
   set_d1(group);

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_h(text_height(font));
   set_w(text_length(font,(unsigned char*)text) + 2*h());
   }
//----------------------------------------------------------------------------
// Destructor
radio_button_object::~radio_button_object (void)
   {
   }

//----------------------------------------------------------------------------
// Keyboard object
//----------------------------------------------------------------------------
// Constructor
keyboard_object::keyboard_object (int(*pro)(void), const int ke)
   :proc_object(d_keyboard_proc,(void**)(&proc)),
	proc(pro),
   keypress(ke)
   {
   // Setup the defaults
   set_x(0);
   set_y(0);
   set_w(0);
   set_h(0);
   set_key(keypress);
   }
//----------------------------------------------------------------------------
// Destructor
keyboard_object::~keyboard_object (void)
   {
   }

//----------------------------------------------------------------------------
// Editable text string object
//----------------------------------------------------------------------------
// Constructor
edittext_object::edittext_object (char *tex, const int len)
   :proc_object(d_edit_proc,(void**)(&text)),
	text(tex),
   allocated(false),
	max_length(len)
   {
   // Deal with making our own memory if we aren't given any
   if(tex==NULL)
      {
      text = new char[len+1];
      strcpy(text,"");
      allocated = true;
      }
   if(text==NULL) degui_no_memory();

 	set_d1(max_length);
	set_d2(strlen(text));

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_h(text_height(font));
   set_w(text_length(font,(unsigned char*)"W") * len);

   // Default colors are slightly different
   set_bg(gui_mg_color);
   }
//----------------------------------------------------------------------------
// Destructor
edittext_object::~edittext_object (void)
   {
   if(allocated==true) delete text;
   }
//----------------------------------------------------------------------------
// Draw the editable text with a line under
void edittext_object::msg_draw (void)
   {
   hline(screen,x(),y()+text_height(font),x()+w(),fg());
   proc_object::msg_draw();
   }
//----------------------------------------------------------------------------
// Set and get the editable text
void edittext_object::set_text (const char *tex)
   {
   if(tex==NULL) return;

   // Check if new length is longer than the old
   int l = strlen(text) + 1;
   if(l>max_length)
      {
      // Expand the room for the new stuff
      if(allocated==true) delete text;
      text = new char[l];
      allocated = true;
      }
   // Copy the stuff over
   strcpy(text,tex);

   redraw();
   }

//----------------------------------------------------------------------------
// List object
//----------------------------------------------------------------------------
// Constructor
list_object::list_object (char*(*proc)(int,int*))
   :proc_object(d_list_proc,(void**)(&list_grabber_proc)),
	list_grabber_proc(proc)
   {
   if(proc==NULL) list_grabber_proc = default_list_object_procedure;

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_h(100);
   set_w(100);
   }
//----------------------------------------------------------------------------
// Destructor
list_object::~list_object (void)
   {
   }

//----------------------------------------------------------------------------
// Default Callback function to specify the contents of a listbox
char *default_list_object_procedure (int index, int *list_size)
   {
   static char *strings[] = {"Zero","One","Two","Three","Four","Five","Six","Seven","Eight","Nine","Ten"};
   if(index<0)
      {
      *list_size = 11;
      return NULL;
      }
   else
      {
      return strings[index];
      }
   }

//----------------------------------------------------------------------------
// Textbox object
//----------------------------------------------------------------------------
// The list procedure to get the text lines
textbox_object *textbox_object::current_textbox;

char* default_textbox_object_text = "This is the default text for a textbox object";

// The number of elements to expand when expanding allocation for line breaks
// Larger number mean faster setup but more memory wasteage
int textbox_object::allocation_step = 10;

//----------------------------------------------------------------------------
// Constructor
textbox_object::textbox_object (char *str, const bool wrap, const bool wwrap)
   :list_object(textbox_object::list_procedure),
   text(NULL),
   length(0),
   self_allocated(false),
   _lines(0),
   wrapping(wrap),
   word_wrapping(wwrap),
   width(9999),
   last_width(-1),
   last_font(NULL),
   places(NULL),
   allocated(0),
   buffered_place(NULL)
   {
   set_text(str);

   // Setup the defaults
   set_x(100);
   set_y(100);
   set_h(100);
   set_w(100);
   }
//----------------------------------------------------------------------------
// Destructor
textbox_object::~textbox_object (void)
   {
   // Unbuffer anything and release the memory
   if(buffered_place!=NULL) buffered_place[0] = buffered_char;
   if(places!=NULL) delete [] places;
   if(self_allocated==true) delete text;
   }
//----------------------------------------------------------------------------
// Helper function to determine line lengths
char* textbox_object::find_line_end (char *string)
   {
   int c = 0;
   char *i = string;

   static char temp[5];
   temp[1] = '\0';

   // Find the line length
   while(1)
      {
      // Check to see if we ran out of string
      if(*i=='\0') return NULL;

      // Check for a DOS newline
      if(*i==13)
         {
         if(*i==10) i++;
         break;
         }
      // Check for a UNIX ??? newline
      if(*i==10) break;

      // If the line too long for the wrapping
      if((c>=width)&&(wrapping==true))
         {
         char *start = i;

         // Line width goes over the end, go back for the whitepsace
         if(word_wrapping==true)
            {
            while(!isspace(*i))
               {
               i--;

               // Check for long lines
               if(i<=string)
                  {
                  i = start;
                  i--;
                  break;
                  }
               }
            i++;
            }
         i--;
         break;
         }

      // Find the width of the character
      temp[0] = *i;
      c += text_length(font,(unsigned char*)temp);

      // Check out the next character
      i++;
      }
   return i;
   }

//----------------------------------------------------------------------------
// Split the text into seperate text portions
void textbox_object::set_text (char *str)
   {
   // Free memory from the last text parse
   if(places!=NULL)
      {
      // Replace any old bufferd chars
      if(buffered_place!=NULL) *buffered_place = buffered_char;
      buffered_place = NULL;

      // Release the old stuff
      if(places!=NULL) delete [] places;
      places = NULL;
      allocated = 0;
      text = NULL;
      length = 0;
      _lines = 0;
      }
   // Make sure we have a reasonable string in our own space
   if(str==NULL)
      {
      text = new char[strlen(default_textbox_object_text)+1];
      if(text==NULL) degui_no_memory();
      strcpy(text,default_textbox_object_text);
      }
   // Or just setup the new text pointers
   else
      {
      text = str;
      }
   length = strlen(text);

   // Make sure we have a reasonable length to deal with
   if(length<2) return;

   // Allocate some initial space for the line breaks
   places = new char*[allocation_step];
   if(places==NULL) degui_no_memory();
   allocated = allocation_step;
   int x;
   for(x=0;x<allocation_step;x++) places[x] = NULL;

   // Get all the line breaks
   char *i = text;
   char *end = text + length;
   while(true)
      {
      // Find the line break
      i = find_line_end(i);

      // If it's a normal line break, remember it
      if(i!=NULL)
         {
         places[_lines] = i+1;
         _lines++;
         }
      // If it's the last line break remember it
      else
         {
         places[_lines] = end;
         _lines++;
         break;
         }

      // Put the stuff in the data arrays
      if(_lines>=allocated)
         {
         // Make new allocation
         char **newplaces = new char*[allocated+allocation_step];
         if(newplaces==NULL) degui_no_memory();

         // Copy the stuff over
         for(x=0;x<allocated;x++)
            {
            newplaces[x] = places[x];
            }
         // Put invalids in the unused but allocated stuff
         for(x=allocated;x<allocated+allocation_step;x++)
            {
            newplaces[x] = NULL;
            }

         // Release the old stuff
         if(places!=NULL) delete [] places;

         // Update the allocation
         places = newplaces;
         allocated += allocation_step;
         }
      // Find the next end point
      i++;
      }
   redraw();
   }
//----------------------------------------------------------------------------
// Getting lines from out of the text string
char* textbox_object::get_line (const int line)
   {
   // Make sure that we have a reasonable request
   if(line>=_lines) return NULL;

   // Replace any old bufferd chars
   if(buffered_place!=NULL) *buffered_place = buffered_char;

   // Buffer the what is under the end of line marker
   buffered_place = places[line];
   buffered_char = *buffered_place;

   // Place the end of line marker
   *buffered_place = '\0';

   // Give up the first line
   if(line==0) return text;

   // Get the line for the user
   return places[line-1];
   }
//---------------------------------------------------------------------------
// Helper function for textbox to get text into lines for listbox
char* textbox_object::list_procedure (int index, int *list_size)
   {
   // Some alais to help readability
   textbox_object *object = textbox_object::current_textbox;

   // Possible readjust the text if the size or font has changed
   // Check if we must reassign the breaks
   if((object->w()!=object->last_width)||(font!=object->last_font))
      {
      // Possibly adjust the width if there is wrapping
      // Assume that there is a scrollbar
      if(object->wrapping==true) object->width = object->w()-30;

      // Reformat the text for this width
      object->set_text(object->text);

      // Check if we need a scrollbar
      if(object->lines() < ((object->h()-3)/text_height(font)))
         {
         // Expand the width for the scrollbar thats not there
         object->width += 10;

         // Reformat the text for this width
         object->set_text(object->text);
         }
      object->last_width = object->w();
      object->last_font = font;
      }

   // Get the text lines or the size
   if(index<0)
      {
      *list_size = object->lines();
      return NULL;
      }
   else
      {
      return object->get_line(index);
      }
   }
//----------------------------------------------------------------------------
// Basic message passing functions just setup the list procedure and call the list object stuff

// Initialization of the dialog
void textbox_object::msg_start (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_start();
   }

// De-initialization of the dialog
void textbox_object::msg_end (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_end();
   }

// Tell the object to draw itself
void textbox_object::msg_draw (void)
   {
   // Make sure nothing gets selected
   set_d1(-1);

   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_draw();

   // Make sure nothing was selected
   set_d1(-1);
   }

// Tell the object to deal with a mouse click default is the same as a keypress
void textbox_object::msg_click (void)
   {
   // Figure out if it's on the text or the scrollbar
   int height = (h()-3) / text_height(font);
   int bar = (_lines > height);

   // Clicked on the text area
   if((!bar)||(mouse_x < x()+w()-10))
      {
      return;
      }
	// Clicked on the scroll area
   else
      {
		// Setup the list getter procedure and pass the message to the list
      current_textbox = this;
      list_object::msg_click();
      }
   }

// Tell the object to deal with a double click
void textbox_object::msg_dclick (void)
   {
   // Ignore any double clicks because we don't want any selection
   }

// Tell the object that a shortcut key was pressed
void textbox_object::msg_key (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_key();
   }

// Tell the object that a key was pressed while it had the focus
// This should return true if the key was used
bool textbox_object::msg_char (const int c)
   {
   // Setup the list getter procedure
   current_textbox = this;

   // This code taken from Allegro's GUI.C with some major mods
   int start = d2();
	if(_lines>0)
      {
      int top, bottom,l;

	   if(d2()>0) top = d2()+1;
	   else top = 0;

      l = (h()-3)/text_height(font);

	   bottom = d2() + l - 1;
	   if(bottom >= _lines-1) bottom = _lines-1;
	   else bottom--;

      if ((c>>8) == KEY_UP)        set_d2(d2()-1);
      else if ((c>>8) == KEY_DOWN) set_d2(d2()+1);
	   else if ((c>>8) == KEY_HOME) set_d2(0);
	   else if ((c>>8) == KEY_END)  set_d2(_lines-l);
	   else if ((c>>8) == KEY_PGUP) set_d2(d2()-(bottom-top));
      else if ((c>>8) == KEY_PGDN) set_d2(d2()+(bottom-top));

      // Make sure that the list stays in bounds
      if(d2()>_lines-l) set_d2(_lines-l);
      if(d2()<0) set_d2(0);
	   }
   else return false;

   /* if we changed something, better redraw... */
   if(d2()!=start)
      {
      show_mouse(NULL);
      msg_draw();
      show_mouse(screen);
      return true;
      }
   return false;
   }

// Tell the object that a key was pressed while it did noty have the focus
// This should return true if the key was used
bool textbox_object::msg_xchar (const int c)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   return list_object::msg_xchar(c);
   }

// Ask the object if they want the focus
bool textbox_object::msg_wantfocus (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   return list_object::msg_wantfocus();
   }

// Tell the object that the got or lost the focus
void textbox_object::msg_gotfocus (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_gotfocus();
   }
void textbox_object::msg_lostfocus (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_lostfocus();
   }

// Tell the object that they got or lost the mouse
void textbox_object::msg_gotmouse (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_gotmouse();
   }
void textbox_object::msg_lostmouse (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_lostmouse();
   }

// Tell the object that the object manager is bored
void textbox_object::msg_idle (void)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_idle();
   }

// Deselect grouped radio buttons
void textbox_object::msg_radio (const int c)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_radio(c);
   }

// Get an hitherto unknown message
void textbox_object::msg_unknown (const int c)
   {
   // Setup the list getter procedure and pass the message to the list
   current_textbox = this;
   list_object::msg_unknown(c);
   }

//----------------------------------------------------------------------------
// Wrapper for MENU struct
//----------------------------------------------------------------------------
// Constructor
menu::menu (void)
	:internal(NULL),
	number(0)
   {
   }
//----------------------------------------------------------------------------
// Destructor
menu::~menu (void)
   {
   }
//----------------------------------------------------------------------------
// Add something to the menu
void menu::_add (const char *text, int(*proc)(void), const MENU *child)
	{
	// Make space for the new stuff
   MENU* new_internal = new MENU[number+2];
   if(new_internal==NULL) degui_no_memory();

	// Copy the old stuff over
	for(int i=0;i<number;i++) new_internal[i] = internal[i];

	// Add the new object
	new_internal[number].text = (char*)text;
	new_internal[number].proc = (int(*)(...))proc;
  	new_internal[number].child = (MENU*)child;

	// Make sure that there is a correct last NULL at the end
	new_internal[number+1].text = NULL;
	new_internal[number+1].proc = NULL;
  	new_internal[number+1].child = NULL;

	// Delete the old stuff
	if(internal!=NULL) delete [] internal;

	// Exchange the pointer sets
	internal = new_internal;
   number++;
	}
//----------------------------------------------------------------------------
// Add a child menu
void menu::add (const char *text, const menu &child)
	{
   _add(text,NULL,child.internal);
	}
//----------------------------------------------------------------------------
// Add a procedure
void menu::add (const char *text, int(*proc)(void))
	{
   _add(text,proc,NULL);
	}

//----------------------------------------------------------------------------
// Menu object
//----------------------------------------------------------------------------
// Constructor
menu_object::menu_object (const menu& men)
   :proc_object(d_menu_proc,(void**)(&internal)),
   internal(men.internal)
   {
   // Setup the defaults
   set_x(0);
   set_y(0);
   set_h(0);
   set_w(0);
   }
//----------------------------------------------------------------------------
// Destructor
menu_object::~menu_object (void)
   {
   }
//----------------------------------------------------------------------------
// Basic message passing functions

// Modification of how the menu works
void menu_object::msg_gotmouse (void)
	{
   }

//----------------------------------------------------------------------------
// Simple object to clear the screen and put up a grid
//----------------------------------------------------------------------------
// Constructor
grid_object::grid_object (const int xg, const int yg)
   :x_gran(xg),
   y_gran(yg)
   {
   // Setup the defaults
   set_x(0);
   set_y(0);
	set_w(0);
   set_h(0);
   }
//----------------------------------------------------------------------------
// Destructor
grid_object::~grid_object (void)
   {
   }
//----------------------------------------------------------------------------
// Basic message passing functions

// Draw the grid object
void grid_object::msg_draw (void)
	{
   // Crear the screen
   set_clip(screen, 0, 0, SCREEN_W-1, SCREEN_H-1);
   clear_to_color(screen, gui_bg_color);

   int x_step = x_gran;
   int y_step = y_gran;

   // Make sure that there is a reasonable grid size
   if((x_step<2)&&(y_step<2)) return;

   // And draw the grid
   int x = 0, y = 0;
   while(x<SCREEN_W)
      {
      while(y<SCREEN_H)
         {
         putpixel(screen,x,y,gui_fg_color);
         y += y_step;
         }
      x+= x_step;
      y = 0;
      }
	}

/*
//----------------------------------------------------------------------------
// An object to hold groups of objects NOTE: Undocumented !!!
//----------------------------------------------------------------------------
// Constructor
dialog_object_group::dialog_object_group (void)
   {
   }
// Constructor
dialog_object_group::dialog_object_group (dialog_object* one, ...)
   {
   // Init the argument list
   va_list args;
   va_start(args,one);

   // Add each of the objects to our internal list
   dialog_object *temp = one;
   while(temp!=NULL)
      {
      objects.add(temp);
      temp=va_arg(args,dialog_object*);
      }
   va_end(args);
   }
//----------------------------------------------------------------------------
// Destructor
dialog_object_group::~dialog_object_group (void)
   {
   }
*/

//----------------------------------------------------------------------------
// Allegro dialog managing class
//----------------------------------------------------------------------------
// Constructor
dialog::dialog (const int fh, const int fw)
	:object_dialogs(NULL),
   new_object_dialogs(NULL),
	number(0),
   restart(false),
   executing(false),
   _close(false),
   _redraw(false),
   font_height(fh),
   font_width(fw)
   {
 	// Add the control object
   add(dialog::dialog_proc,0,0,0,0,0,0,0,0,0,0,this);
   }
//----------------------------------------------------------------------------
// Destructor
dialog::~dialog (void)
   {
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   // Give back DIALOG *d to the still existing objects
	if(list!=NULL)
      {
      for(int i=0;i<number;i++)
         {
         if(list[i].proc==dialog_object::object_message_proc)
            {
            // Find the object
            dialog_object *object = (dialog_object*)list[i].dp;

            // Give back the d pointer and any init stuff
            object->disconnect();
            object->_parent = NULL;
            }
         }
      delete [] list;
      }
   }
//----------------------------------------------------------------------------
// Add a procedure object to the dialog
void dialog::add (int(*proc)(int,DIALOG*,int), const int x,  const int y, const int w, const int h, const int fg, const int bg, const int key, const int flags, const int d1, const int d2, void *dp)
	{
   // Make space for the new stuff plus space for the standard dialogs
   DIALOG *temp = new DIALOG[number+3];
   if(temp==NULL) degui_no_memory();

	// Copy the old stuff over
	for(int i=0;i<number;i++)
      {
      if(new_object_dialogs!=NULL) temp[i] = new_object_dialogs[i];
      else temp[i] = object_dialogs[i];
      }

	// Add the new object
	temp[number].proc = proc;
	temp[number].x = x;
	temp[number].y = y;
	temp[number].w = w;
	temp[number].h = h;
	temp[number].fg = fg;
	temp[number].bg = bg;
	temp[number].key = key;
	temp[number].flags = flags;
	temp[number].d1 = d1;
	temp[number].d2 = d2;
   temp[number].dp = dp;

	// Make sure that there is a correct last NULL at the end
	temp[number+1].proc = NULL;
	temp[number+1].x = 0;
	temp[number+1].y = 0;
	temp[number+1].w = 0;
	temp[number+1].h = 0;
	temp[number+1].fg = 0;
	temp[number+1].bg = 0;
	temp[number+1].key = 0;
	temp[number+1].flags = 0;
	temp[number+1].d1 = 0;
	temp[number+1].d2 = 0;
	temp[number+1].dp = NULL;

	// Delete the old stuff
	// Exchange the pointer sets
	if(new_object_dialogs!=NULL) delete [] new_object_dialogs;
   new_object_dialogs = temp;
   number++;

   // If we are running add the object
   reset();
	}
//----------------------------------------------------------------------------
// Add an object to the dialog
void dialog::add (dialog_object &obj, const int x, const int y, const int w, const int h, const int fg, const int bg)
   {
   // Figure out the object's previous state
	int key = obj.key();
   int flags = obj.flags();
   int d1 = obj.d1();
   int d2 = obj.d2();

   int realx;
   if(x>=0) realx = x;
   else realx = obj.x();
   obj.set_x(realx);

   int realy;
   if(y>=0) realy = y;
   else realy = obj.y();
   obj.set_y(realy);

   int realw;
   if(w>=0) realw = w;
   else realw = obj.w();
   obj.set_w(realw);

   int realh;
   if(h>=0) realh = h;
   else realh = obj.h();
   obj.set_h(realh);

   int realfg;
   if(fg>=0) realfg = fg;
   else realfg = obj.fg();
   obj.set_fg(realfg);

   int realbg;
   if(bg>=0) realbg = bg;
   else realbg = obj.bg();
   obj.set_bg(realbg);

   // Set the objects parent dialog
   obj._parent = this;

   // Add the C++ object to the dialog
   add(dialog_object::object_message_proc,realx,realy,realw,realh,realfg,realbg,key,flags,d1,d2,&obj);
	}
//----------------------------------------------------------------------------
// Do the dialog
dialog_object* dialog::execute (const dialog_object* focus_object)
	{
   // Prevent reentrancy
	if(executing==true) return NULL;
   executing = true;

   // Make sure the desired focus object is in our dialog
   int focus_obj = -1;
   if(new_object_dialogs!=NULL) object_dialogs = new_object_dialogs;
   new_object_dialogs = NULL;
 	for(int i=0;i<number;i++)
      {
      if(object_dialogs[i].dp==focus_object) focus_obj = i;
      }

	// Just ask allegro to do the dialog
   do {
      restart = false;

      // Get the newest object list if there is one
      if(new_object_dialogs!=NULL) object_dialogs = new_object_dialogs;
      new_object_dialogs = NULL;

      // Ask Allegro to do the dialog
      focus_obj = do_dialog(object_dialogs,focus_obj);
      }
      // If an object caused a restart, then do the dialog again
      while(restart==true);

   // Cleanup the stuff
   new_object_dialogs = object_dialogs;
   object_dialogs = NULL;

   // Return the dialog that exited
   executing = false;
   if(focus_obj>=0) return (dialog_object*)(new_object_dialogs[focus_obj].dp);
   else return NULL;
	}
//----------------------------------------------------------------------------
// Do the dialog as a popup
dialog_object* dialog::popup (const dialog_object* focus_object)
	{
   // Find the extrema of the dialog
   int minx = MAXINT;
   int maxx = -MAXINT;
   int miny = MAXINT;
   int maxy = -MAXINT;
   find_extrema(0,number,minx,miny,maxx,maxy);

   // Get the background
   BITMAP *background = create_bitmap(maxx-minx+1,maxy-miny+1);
   show_mouse(NULL);
   blit(screen,background,minx,miny,0,0,maxx-minx+1,maxy-miny+1);
   show_mouse(screen);

   // Do the dialog
   dialog_object *ret = execute(focus_object);

   // Replace and release the background
   show_mouse(NULL);
   blit(background,screen,0,0,minx,miny,maxx-minx+1,maxy-miny+1);
   show_mouse(screen);
   destroy_bitmap(background);

   // Return the dialog that exited
   return ret;
	}

//----------------------------------------------------------------------------
// Restart the dialog
void dialog::reset (void)
	{
   restart = true;
	}
//----------------------------------------------------------------------------
// Restart the dialog
void dialog::close (void)
	{
   _close = true;
	}
//----------------------------------------------------------------------------
// An Allegro object to get basic messages
// This is a C++ static function
int dialog::dialog_proc (int msg, DIALOG *d, int)
    {
    // Get the dialog object
    dialog *dia = (dialog*)d->dp;

    // Check for a restart or close and exit this dialog
    // and maybe start another one
    if((dia->restart==true)||(dia->_close==true)) return D_CLOSE;

    // Possible redraw
    if(dia->_redraw==true)
       {
       dia->_redraw = false;
       return D_REDRAW;
       }

    switch(msg)
       {
       case(MSG_DRAW):
          {
          break;
          }
       default:
          {
          break;
          }
       }
    return D_O_K;
    }

//----------------------------------------------------------------------------
// Redraw the dialog immediately or just request a redraw
void dialog::redraw (const bool immediate)
	{
   // Just request a redraw to be done later
   if(immediate==false)
      {
      _redraw = true;
      }
   // Force a compete redraw of all objects right now
   else
      {
      // Send a redraw to all the objects if we are running
      if(object_dialogs!=NULL)
         {
         show_mouse(NULL);
         for(int i=0;i<number;i++)
            {
            object_dialogs[i].proc(MSG_DRAW,&object_dialogs[i],0);
            }
         show_mouse(screen);
         }
      }
	}
//----------------------------------------------------------------------------
// Helper function to find the total space of a dialog
void dialog::find_extrema (const int start, const int end, int &minx, int &miny, int &maxx, int &maxy)
   {
   minx = MAXINT;
   maxx = -MAXINT;
   miny = MAXINT;
   maxy = -MAXINT;

   // Find out how many objects to center
   int num = number;
   if(end<num) num = end;

   // Get the newest object list if there is one
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   // Adjust the size for the font's as necessary
   scale_to_font(font);

   // Find the extremes of the dialog, but only for reasonable objects
   int x, y, w, h;
   for(int i=start;i<num;i++)
      {
      // Find the object position
      x = list[i].x;
      y = list[i].y;
      w = list[i].w;
      h = list[i].h;

      // Find the extrema for real objects
      if((w>0)&&(h>0))
         {
         if(x<minx) minx = x;
         if((x+w)>maxx) maxx = x+w;
         if(y<miny) miny = y;
         if((y+h)>maxy) maxy = y+h;
         }
      }
   }
//----------------------------------------------------------------------------
// Center the dialog
void dialog::center (const int start, const int end)
	{
   // Find the extrema of the dialog
   int minx = MAXINT;
   int maxx = -MAXINT;
   int miny = MAXINT;
   int maxy = -MAXINT;
   find_extrema(start,end,minx,miny,maxx,maxy);

   // Find the center of all the parts
   int cx = minx+((maxx-minx)/2);
   int cy = miny+((maxy-miny)/2);

   // Find how much the dialogs must be shifted
   int sx = (SCREEN_W/2)-cx;
   int sy = (SCREEN_H/2)-cy;

   // Find out how many objects to center
   int num = number;
   if(end<num) num = end;

   // Shift all the dialog objects
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   for(int i=start;i<num;i++)
      {
      // Move all procedures
      list[i].x += sx;
      list[i].y += sy;

      // Move all objects
      if(list[i].proc==dialog_object::object_message_proc)
         {
         // Get the object
         dialog_object *object = (dialog_object*)list[i].dp;

         // Move it
         object->set_x(object->x()+sx);
         object->set_y(object->y()+sy);
         }
      }
   reset();
	}
//----------------------------------------------------------------------------
// Use a palette while preserving the main gui colours
void dialog::use_palette (PALETTE user_pal)
   {
   // Remember the old colors
   RGB *old_fg_colors = new RGB[number];
   if(old_fg_colors==NULL) degui_no_memory();
   RGB *old_bg_colors = new RGB[number];
   if(old_bg_colors==NULL) degui_no_memory();
   RGB old_gui_fg_color;
   RGB old_gui_mg_color;
   RGB old_gui_bg_color;

   // Get the old RGB values
   DIALOG* current = object_dialogs;
   if(new_object_dialogs!=NULL) current = new_object_dialogs;
   int i;
   for(i=0;i<number;i++)
      {
      get_color(current[i].fg,&(old_fg_colors[i]));
      get_color(current[i].bg,&(old_bg_colors[i]));
      }
   get_color(gui_fg_color,&old_gui_fg_color);
   get_color(gui_mg_color,&old_gui_mg_color);
   get_color(gui_bg_color,&old_gui_bg_color);

   // Now set the new palette
   set_palette(user_pal);

   // Find the closest colours for this palette
   gui_fg_color = makecol8(old_gui_fg_color.r*4,old_gui_fg_color.g*4,old_gui_fg_color.b*4);
   gui_mg_color = makecol8(old_gui_mg_color.r*4,old_gui_mg_color.g*4,old_gui_mg_color.b*4);
   gui_bg_color = makecol8(old_gui_bg_color.r*4,old_gui_bg_color.g*4,old_gui_bg_color.b*4);

   // Force the closest colors to be exactly what we want
   vsync();
   set_color(gui_fg_color,&old_gui_fg_color);
   set_color(gui_mg_color,&old_gui_mg_color);
   set_color(gui_bg_color,&old_gui_bg_color);

   // Adjust the rest of the colour to the nearest RGB values
   if(new_object_dialogs!=NULL)
      {
      for(i=0;new_object_dialogs[i].proc;i++)
         {
         new_object_dialogs[i].fg = makecol8(old_fg_colors[i].r*4,old_fg_colors[i].g*4,old_fg_colors[i].b*4);
         new_object_dialogs[i].bg = makecol8(old_bg_colors[i].r*4,old_bg_colors[i].g*4,old_bg_colors[i].b*4);
         }
      }
   if(object_dialogs!=NULL)
      {
      for(i=0;object_dialogs[i].proc;i++)
         {
         object_dialogs[i].fg = makecol8(old_fg_colors[i].r*4,old_fg_colors[i].g*4,old_fg_colors[i].b*4);
         object_dialogs[i].bg = makecol8(old_bg_colors[i].r*4,old_bg_colors[i].g*4,old_bg_colors[i].b*4);
         }
      }
   delete [] old_fg_colors;
   delete [] old_bg_colors;

   // Do a redraw
   redraw();
   }
//----------------------------------------------------------------------------
// Scale the dialog for a particular font
void dialog::scale_to_font (FONT *font)
   {
   DIALOG *list = object_dialogs;
   if(new_object_dialogs!=NULL) list = new_object_dialogs;

   float height_scale = (float)font_height/(float)text_height(font);
   float width_scale = (float)font_width/(float)text_length(font,(unsigned char*)"H");
   for(int i=0;i<number;i++)
      {
      list[i].x = (int)((float)list[i].x / width_scale);
      list[i].y = (int)((float)list[i].y / height_scale);
      list[i].w = (int)((float)list[i].w / width_scale);
      list[i].h = (int)((float)list[i].h / height_scale);
      }
   font_height = text_height(font);
   font_width = text_length(font,(unsigned char*)"H");
   }

//----------------------------------------------------------------------------
// Some global error handling procedures
//----------------------------------------------------------------------------
// General error handling procedure
void degui_error (const char *message)
   {
   allegro_exit();
   fprintf(stderr,"%s\n",message);
   exit(1);
   }
//----------------------------------------------------------------------------
// Error from a failed new
void degui_no_memory (void)
   {
   degui_error("DEGUI: No memory, new has failed");
   }
//----------------------------------------------------------------------------



