//----------------------------------------------------------------------------
//
// DEGUI - C++ Objects for Allegro's gui
//
// Douglas Eleveld (D.J.Eleveld@anest.azg.nl)
//
//----------------------------------------------------------------------------

/*
  INI-File class
  -------------------------------------------------------------
  (c) 1997 M.Baas
*/

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include "inifile.h"


/*==================================================================
                              INI file
==================================================================*/

/*-------------------------------open-----------------------------
  Open INI file (that is, connect a file with the object).
-----------------------------------------------------------------*/
bool INIfile::open(char *filename)
{
  ifstream f(filename);
  if (!f.good()) return false;
  char s[500];
  int          line=0;
  commentRow   comment;
  parameterRow parameter;
  sectionRow   section;
  row*         r;

  // Just to be sure that the list is empty.
  rows.free_list();
  sections.clear();
  changed=false;

  // Dummy section for parameters at the beginning without a section
  sectionRow *srow = new sectionRow(0);
  if (srow==NULL) return false;
  char empty[3]="[]";
  if (!srow->set_string(empty)) { delete srow; return false; };
  rows.insert_last(srow);

  // Read INI file...
  while(!f.eof())
  {
    // Read line
    f.getline(s,sizeof(s));
    if (f.eof() && *s==0) break;
    line++;
    // Is it a comment?
    if (comment.is_comment(s)) r = new commentRow(line);
    else
    {
      // Is it a parameter?
      if (parameter.is_parameter(s)) r = new parameterRow(line);
      else
      {
        // A section?
        if (section.is_section(s)) r = new sectionRow(line);
        else
          // What then? Well, we treat it as comment...
          r = new commentRow(line);
      };
    };
    if (r==NULL) return false;
    if (!r->set_string(s)) { delete r; return false; };
    rows.insert_last(r);
  };

  // Initialize sections & build section lookup table...
  r=(row*)rows.first;
  while(r!=NULL)
  {
    // Section? then call setup()
    if (r->row_type() == row::section)
    {
      sectionRow* sr = (sectionRow*)r;
      // Call setup()
      if (!sr->setup()) return false;
      // Add section to the lookup table
      if (!sections.add(sr)) return false;
      // Continue after the last parameter (if there is one)
      if (sr->lastParam != NULL) r = sr->lastParam;
    };
    if (r!=NULL) r=(row*)r->succ;
  };
  // Sort the lookup table
  sections.sort();

  // Copy filename...
  name=(char*)malloc(strlen(filename)+1);
  if (name==NULL) return false;
  strcpy(name,filename);

  return true;
}


/*------------------------------create----------------------------
  Create a new INI file.
-----------------------------------------------------------------*/
bool INIfile::create(char *filename)
{
  // Just to be sure that the list is empty.
  rows.free_list();
  sections.clear();
  changed=false;

  // Dummy section for parameters at the beginning without a section
  sectionRow *srow = new sectionRow(0);
  if (srow==NULL) return false;
  char empty[3]="[]";
  if (!srow->set_string(empty)) { delete srow; return false; };
  rows.insert_last(srow);
  // Call setup()
  if (!srow->setup()) return false;
  // Add section to the lookup table
  if (!sections.add(srow)) return false;

  // Copy filename...
  name=(char*)malloc(strlen(filename)+1);
  if (name==NULL) return false;
  strcpy(name,filename);

  changed=true;
  return true;
}

/*-------------------------------close----------------------------
-----------------------------------------------------------------*/
bool INIfile::close()
{
  if (changed && name!=NULL)
  {
    ofstream of(name);
    if (!of.good) return false;

    // First row is the dummy object
    row *r = (row*)rows.first->succ;
    while(r!=NULL)
    {
      r->write(of);
      r=(row*)r->succ;
    };
    changed=false;
  };

  sections.clear();
  rows.free_list();
  free(name);
  name=NULL;
  return true;
}

/*----------------------------get_parameter-----------------------
-----------------------------------------------------------------*/
char* INIfile::get(char *section, char *param)
{
  int index,n;

  // Look for the section in the lookup table
  if (sections.index_of(section,index,n))
  {
     // Search the parameter in all sections with the given name...
     for(int i=0; i<n; i++)
     {
       sectionRow *sr   = (sectionRow*)sections[index+i];
       char*      value = sr->get_parameter(param);
       if (value!=NULL) return value;
     };
  };

  return NULL;
}

/*----------------------------set_parameter-----------------------
-----------------------------------------------------------------*/
bool INIfile::set(char *section, char *param, char *value)
{
  int index,n,res;
  sectionRow *sr;

  // Look for the section in the lookup table
  if (sections.index_of(section,index,n))
  {
     // Try to set the new value...
     for(int i=0; i<n; i++)
     {
       sr  = (sectionRow*)sections[index+i];
       res = sr->set_parameter(param,value,false,NULL);
       if (res==-1) return false;
       if (res==0)  { changed=true; return true; };
     };
     // Value couldn't be set, this means it doesn't already exist,
     // so we have to create a new one.
     sr = (sectionRow*)sections[index];
     if (sr->set_parameter(param,value,true,&rows)==0)
     {
       changed=true;
       return true;
     }
     else return false;
  }
  else  // Create a new section...
  {
    // Create a section definition
    char* s = (char*)malloc(strlen(section)+3);
    if (s==NULL) return false;
    strcpy(s,"[");
    strcat(s,section);
    strcat(s,"]");

    // Allocate a new sectionRow object
    sr = new sectionRow(0);   // row number isn't important anymore
    if (sr==NULL) { free(s); return false; };

    // Set created definition
    if (!sr->set_string(s))
    {
      free(s);
      delete(sr);
      return false;
    };
    free(s);

    // Insert new section in row list
    rows.insert_last(sr);
    // Update lookup table
    sections.insert_at(sr,index);

    // Add parameter
    if (sr->set_parameter(param,value,true,&rows)==0)
    {
       changed=true;
       return true;
    }
    else return false;
  };
}

/*------------------------get_section_name------------------------
-----------------------------------------------------------------*/
char* INIfile::get_section_name(int section_nr)
{
  row* r=sections[section_nr];
  if (r!=NULL)
     return r->get_identifier();
  else
     return NULL;
}

/*-------------------------get_parameter_name---------------------
-----------------------------------------------------------------*/
char* INIfile::get_parameter_name(int section_nr, int param_nr)
{
  sectionRow *sr=(sectionRow*)sections[section_nr];
  row *r=sr->params[param_nr];
  if (r!=NULL)
     return r->get_identifier();
  else
     return NULL;
}

/*-------------------------get_parameter--------------------------
  Read an char string value with default
-----------------------------------------------------------------*/
char* INIfile::get(char *section, char *param, char *c)
{
  char* ret = c;
  char *s = get(section,param);
  if (s!=NULL) ret = s;
  return ret;
}

/*-------------------------get_parameter--------------------------
  Read an integer value with default
-----------------------------------------------------------------*/
int INIfile::get(char *section, char *param, int i)
{
  int ret = i;
  char *s = get(section,param);
  if (s!=NULL) ret=atoi(s);
  return ret;
}

/*-------------------------get_parameter--------------------------
  Read a double value with default
-----------------------------------------------------------------*/
double INIfile::get(char *section, char *param, double d)
{
  double ret = d;
  char *s = get(section,param);
  if (s!=NULL) d=atof(s);
  return ret;
}

/*-------------------------get_parameter--------------------------
  Read a boolean value with default
-----------------------------------------------------------------*/
bool INIfile::get(char *section, char *param, bool b)
{
  bool ret = b;
  char *s = get(section,param);
  if (s!=NULL) ret = (stricmp(s,"TRUE")==0);
  return ret;
}


/*-------------------------set_parameter--------------------------
  Write an integer value.
-----------------------------------------------------------------*/
bool INIfile::set(char *section, char *param, const int i)
{
  char s[40];
  itoa(i,s,10);
  return set(section,param,s);
}

/*-------------------------set_parameter--------------------------
  Write a double value.
-----------------------------------------------------------------*/
bool INIfile::set(char *section, char *param, const double d)
{
  char s[40];
  sprintf(s,"%f",d);
  return set(section,param,s);
}

/*-------------------------set_parameter--------------------------
  Write a boolean value.
-----------------------------------------------------------------*/
bool INIfile::set(char *section, char *param, const bool b)
{
  char s[10] = "False";
  if (b) strcpy(s,"True");
  return set(section,param,s);
}

/*==================================================================
                            Lookup table
==================================================================*/

/*--------------------------insert_at------------------------------
  Insert row r at the given index.

  Pre: index must be in a valid range (0,1,...,number).
  Input:  r     - Row to be inserted into the table
          index - This will be the new index of r
  Return: False, if there was not enough memory to create a new table
-----------------------------------------------------------------*/
bool lookupTable::insert_at(row* r,int index)
{
  // Do we have to enlarge the table?
  if (number==tableSize)
  {
    int  newsize  = (tableSize==0)? 1 : 2*tableSize;
    // Allocate the new table...
    row** newtable = new row* [newsize];
    if (newtable==NULL) return false;
    // Copy the elements before index...
    for(int i=0; i<index; i++)
      newtable[i]=table[i];
    // Copy the elements after index...
    for(int i=index; i<number; i++)
      newtable[i+1]=table[i];
    // Insert new row
    newtable[index]=r;
    number++;
    // Keep the new table and delete the old one
    delete [] table;
    table     = newtable;
    tableSize = newsize;
  }
  else   // There's no need for a new table...
  {
    // Copy elements after index...
    for(int i=number-1; i>=index; i--)
      table[i+1]=table[i];
    // Insert new row
    table[index]=r;
    number++;
  };

  return true;
}

/*-------------------------------add------------------------------
  Add a new row to the table.
  The resulting table isn't sorted yet!

  Input:  r     - Row to be inserted into the table
  Return: False, if there was not enough memory to create a new table
-----------------------------------------------------------------*/
bool lookupTable::add(row* r)
{
  return insert_at(r,number);
}


/*-----------------------------index_of---------------------------
  Search a row with identifier k.

  Pre:    The table must be sorted!
  Input:  k     -  Identifier to search
  Output: index -  Index of first row with identifier k
                   or index where k should be inserted to maintain
                   the sorted order of the table.
          n     -  Number of rows with identifier k
  Return: True, if a row was found.
-----------------------------------------------------------------*/
bool lookupTable::index_of(char* k, int &index, int &n)
{
  int l = 0;
  int r = number-1;
  int cmpres;

  // Searching row with identifier k ...
  while(l<=r)
  {
    index  = (l+r)/2;
    cmpres = strcmp(table[index]->get_identifier(),k);
    // Identifiers found?
    if (cmpres==0)
    {
      // First row found?
      if (l==r)
      {
        n=1;
        l++;
        // Count the elements with identical identifiers...
        while(l<number)
        {
          if(strcmp(table[l]->get_identifier(),k)!=0) break;
          l++;
          n++;
        };
        return true;
      }
      else r=index;
    }
    else   // Identifier not found yet
    {
      if (cmpres<0) l=index+1; else r=index-1;
    };
  };

  // No row found with identifier k
  index=l;
  n=0;
  return false;
}


static int compareRows(row** r1,row** r2)
{
  int res=strcmp((*r1)->get_identifier(),(*r2)->get_identifier());
  if (res!=0) return res;
  else
  {
    return (*r1)->rowNumber - (*r2)->rowNumber;
  };
}

void lookupTable::sort()
{
  if (number>0) qsort(table,number,sizeof(row*),compareRows);
}

/*--------------------------- [] operator ------------------------
  Return a pointer to the row at the given index.
  If index is not valid the function returns NULL.
-----------------------------------------------------------------*/
row* lookupTable::operator[](int index)
{
  if (index<0 || index>=number)
     return NULL;
  else
     return table[index];
}


/*==================================================================
                                Row
==================================================================*/


/*---------------------------set_string---------------------------
  Set the string that contains the actual INI file row.

  Pre: s!=NULL
-----------------------------------------------------------------*/
bool row::set_string(char *s)
{
  int len=strlen(s);

  // Do we need a larger buffer?
  if (len >= bufferSize)
  {
     free(rowBuffer);
     bufferSize=len+1;
     rowBuffer=(char*)malloc(bufferSize);
     if (rowBuffer==NULL) { bufferSize=0; return false; };
  };
  // Copy the string
  strcpy(rowBuffer,s);
  return true;
};


/*==================================================================
                           parameterRow
==================================================================*/


/*---------------------------set_string---------------------------
  Set the string that contains the actual INI file row which
  must contain a valid parameter.

  Pre: s!=NULL
       s must be a valid parameter row (is_parameter(s) must return true)
-----------------------------------------------------------------*/
bool parameterRow::set_string(char *s)
{
  use_newvalue=false;

  // Copy string
  if (!row::set_string(s)) return false;

  param     = rowBuffer;
  value     = strchr(rowBuffer,'=');
  param_end = value;

  // Delete whitespace at the end of the parameter value
  char *p  = value+strlen(value)-1;
  char *pf = p;
  while(isspace(*p)) p--;
  p[1]=0;
  // Skip whitespace at the beginning of the parameter value
  value++;
  while(isspace(*value)) value++;

  // New values larger than that need a memory allocation
  maxvaluesize_without_allocating = pf-value+1;

  // Skip whitespace at the beginning of the parameter name
  while(isspace(*param)) param++;
  // If param is equal to param_end here then param = ""
  if (param!=param_end)
  {
    // Skip whitespace at the end of the parameter name
    param_end--;
    while(isspace(*param_end)) param_end--;
    param_end++;
  };

  dummy_char = *param_end;
  *param_end = 0;
  return true;
}

/*---------------------------set_value----------------------------
  Set a new parameter value.

  Pre: A parameter was set (=set_string() was called)
       v != NULL
  Return: False, if the new value couldn't be set (no memory)
-----------------------------------------------------------------*/
bool parameterRow::set_value(char *v)
{
  int len = strlen(v);

  // Can the value be stored in the previous allocated memory?
  // newvalue_size has to be -1 if there's no memory allocated!
  if (len<=newvalue_size)
  {
    strcpy(newvalue,v);
    use_newvalue=true;
    return true;
  };

  // Can the value be stored in the original row?
  if (len<=maxvaluesize_without_allocating)
  {
    strcpy(value,v);
    use_newvalue=false;
    return true;
  };

  // v is bigger than all the previous values, so we must allocate new memory

  free(newvalue);
  newvalue=(char*)malloc(len+1);
  if (newvalue==NULL)
  {
    use_newvalue=false;
    return false;
  };
  strcpy(newvalue,v);
  use_newvalue=true;
  return true;
}

/*---------------------------is_parameter-------------------------
  Tests if string s is a valid parameter.
  s is valid if it contains a '='.

  Pre: s!=NULL
-----------------------------------------------------------------*/
bool parameterRow::is_parameter(char *s)
{
  if (strchr(s,'=')!=NULL) return true; else return false;
}

/*-------------------------------write----------------------------
  Writes the row to os.

  Pre: A string was set with set_string()
-----------------------------------------------------------------*/
void parameterRow::write(ostream &os)
{
  *param_end=dummy_char;

  if (use_newvalue)
  {
    char tmp = *value;
    *value=0;
    os<<rowBuffer<<newvalue<<"\n";
    *value=tmp;
  }
  else
  {
    row::write(os);
  };
  *param_end=0;
}


/*==================================================================
                            sectionRow
==================================================================*/

/*---------------------------set_string---------------------------
  Set the string that contains the actual INI file row which
  must contain a valid section.

  Pre: s!=NULL
       s must be a valid section row (is_section(s) must return true)
-----------------------------------------------------------------*/
bool sectionRow::set_string(char *s)
{
  // Copy string
  if (!row::set_string(s)) return false;

  sectionName = rowBuffer+1;
  sectionEnd  = strchr(rowBuffer,']');
  *sectionEnd = 0;
  return true;
}

/*---------------------------is_section---------------------------
  Tests if string s is a valid section.

  Pre: s!=NULL
-----------------------------------------------------------------*/
bool sectionRow::is_section(char *s)
{
  if (*s!='[') return false;
  if (strchr(s,']')==NULL) return false;
  return true;
}

/*-------------------------------write----------------------------
  Writes the row to os.

  Pre: A string was set with set_string()
-----------------------------------------------------------------*/
void sectionRow::write(ostream &os)
{
  *sectionEnd=']';
  row::write(os);
  *sectionEnd=0;
}


/*------------------------------setup-----------------------------
  Initializing parameter lookup table and other stuff.

  Pre: Chained list with rows must be complete.
-----------------------------------------------------------------*/
bool sectionRow::setup()
{
  // Clear lookup table
  params.clear();

  firstParam=lastParam=NULL;

  // Start with the row after this section
  row *r = (row*)succ;

  // Collect parameter definitions...
  while(r!=NULL)
  {
    // Have we reached the top of another section? then we are finished
    if (r->row_type() == row::section) break;
    // Parameter?
    if (r->row_type() == row::parameter)
    {
      // Add the parameter to the lookup table
      if (!params.add(r)) { params.clear(); return false; };
      // Update firstParam/lastParam
      if (firstParam==NULL) firstParam=(parameterRow*)r;
      lastParam=(parameterRow*)r;
    };
    // Next row
    r=(row*)r->succ;
  };
  // Sorting the lookup table
  params.sort();
  return true;
}


/*--------------------------get_parameter-------------------------
  Search a parameter and return its value or NULL if the parameter
  doesn't exist.

  Pre: The lookup table must be initialized (setup() must be called)
-----------------------------------------------------------------*/
char* sectionRow::get_parameter(char *paramname)
{
  int index,n;

  // Look for the parameter in the lookup table
  if (params.index_of(paramname,index,n))
     return ((parameterRow*)params[index])->get_value();
  else
     return NULL;
}


/*--------------------------set_parameter-------------------------
  Input:   paramname - Parameter name
           value     - New value to be set
           create    - True = Create parameter if it doesn't exist
           lst       - List with rows to insert new parameter
  Return:  0 = OK. Value changed/parameter created.
          -1 = No memory to change or create value.
           1 = Parameter not found (only when create=false).
-----------------------------------------------------------------*/
int sectionRow::set_parameter(char *paramname,char *value,bool create,list *lst)
{
  int index,n;

  // Look for the parameter in the lookup table
  if (params.index_of(paramname,index,n))
  {
     // Set the new value
     if (((parameterRow*)params[index])->set_value(value))
        return 0;
     else
        return -1;
  }
  else
  {
     // Should a new parameter be created?
     if (create)
     {
       // Create a parameter definition
       char* s = (char*)malloc(strlen(paramname)+1+strlen(value)+1);
       if (s==NULL) return -1;
       strcpy(s,paramname);
       strcat(s,"=");
       strcat(s,value);

       // Allocate a new parameterRow object
       parameterRow *pr = new parameterRow();
       if (pr==NULL) { free(s); return -1; };

       // Set created definition
       if (!pr->set_string(s)) { free(s); delete(pr); return -1; };
       free(s);

       // Insert new parameter in row list
       if (lastParam!=NULL)
       {
         lst->insert_after(pr,lastParam);
         lastParam=pr;
         if (firstParam==NULL) firstParam=pr;
       }
       else
       {
         lst->insert_after(pr,this);
         firstParam=pr;
         lastParam=pr;
       };

       // Update lookup table
       params.insert_at(pr,index);
       return 0;
     }
     else return 1;
  };
}

